"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import abc
import ast

from abc import ABCMeta
from collections import OrderedDict
from enum import Enum
from nagini_contracts.io_contracts import BUILTIN_IO_OPERATIONS
from nagini_translation.lib.constants import (
    BOXED_PRIMITIVES,
    CALLABLE_TYPE,
    END_LABEL,
    ERROR_NAME,
    INTERNAL_NAMES,
    MODULE_VARS,
    PMSET_TYPE,
    PRIMITIVE_INT_TYPE,
    PRIMITIVE_PREFIX,
    PRIMITIVE_SEQ_TYPE,
    PRIMITIVE_SET_TYPE,
    PRIMITIVES,
    PSEQ_TYPE,
    PSET_TYPE,
    RESULT_NAME,
    STRING_TYPE,
    VIPER_KEYWORDS,
)
from nagini_translation.lib.io_checkers import IOOperationBodyChecker
from nagini_translation.lib.typedefs import Expr, Stmt
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.util import (
    get_column,
    InvalidProgramException,
    SingletonFreshName,
)
from nagini_translation.lib.views import (
    CombinedDict,
    IOOperationContentDict,
)
from typing import Any, Dict, List, Optional, Set, Tuple
from toposort import toposort_flatten


class ContainerInterface(metaclass=ABCMeta):
    """
    Interface implemented by PythonNodes that can contain other PythonNodes,
    and by the translation context. Enables others to get the context of
    this container.
    """
    @abc.abstractmethod
    def get_contents(self, only_top: bool) -> Dict:
        """
        Returns the elements that can be accessed from this container (to be
        used by get_target). If 'only_top' is true, returns only top level
        elements that can be accessed without a receiver.
        """


class PythonScope:
    """
    Represents a namespace/scope in Python
    """
    def __init__(self, sil_names: Set[str], superscope: 'PythonScope'):
        self.sil_names = sil_names
        self.superscope = superscope
        self._module = None

    def contains_name(self, name: str) -> bool:
        if self.sil_names is not None:
            result = name in self.sil_names
        else:
            result = self.superscope.contains_name(name)
        return result

    def get_fresh_name(self, name: str) -> str:
        if self.contains_name(name):
            counter = 0
            new_name = name + '_' + str(counter)
            while self.contains_name(new_name):
                counter += 1
                new_name = name + '_' + str(counter)
            self._add_name(new_name)
            return new_name
        else:
            self._add_name(name)
            return name

    def _add_name(self, new_name: str) -> None:
        if self.sil_names is not None:
            self.sil_names.add(new_name)
        else:
            self.superscope._add_name(new_name)

    @property
    def scope_prefix(self) -> List[str]:
        if self.superscope is None:
            return [self.name]
        else:
            return self.superscope.scope_prefix + [self.name]

    @property
    def module(self) -> 'PythonModule':
        if self._module is not None:
            return self._module
        if self.superscope is not None:
            return self.superscope.module
        else:
            return self


class PythonStatementContainer:
    """
    Class to be mixed into any node that contains executable statements (currently methods
    and modules). Stores the analyzer information related to those statements.
    """
    def __init__(self):
        self.labels = [END_LABEL]
        self.precondition = []
        self.postcondition = []
        self.try_blocks = []  # direct
        self.loop_invariants = {}   # type: Dict[Union[ast.While, ast.For], List[ast.AST]]


class PythonModule(PythonScope, ContainerInterface, PythonStatementContainer):
    def __init__(self, types: TypeInfo,
                 node_factory: 'ProgramNodeFactory',
                 type_prefix: str,
                 global_module: 'PythonModule',
                 node: ast.Module,
                 sil_names: Set[str] = None,
                 file: str = None) -> None:
        """
        Represents a module, i.e. either a directory that only contains other
        modules or an actual file that may contain classes, methods etc.

        :param type_prefix: The prefix identifying this module in TypeInfo.
        :param global_module: The module containing globally available elements.
        :param sil_names: Set of all used Silver names, shared between modules.
        """
        if sil_names is None:
            sil_names = set(VIPER_KEYWORDS + INTERNAL_NAMES)
            sil_names |= set([RESULT_NAME, ERROR_NAME, END_LABEL])
        PythonScope.__init__(self, sil_names, None)
        PythonStatementContainer.__init__(self)
        self.node = node
        self.classes = OrderedDict()
        self.functions = OrderedDict()
        self.methods = OrderedDict()
        self.predicates = OrderedDict()
        self.io_operations = OrderedDict()
        self.global_vars = OrderedDict()
        self.namespaces = OrderedDict()
        self.global_module = global_module
        self.type_prefix = type_prefix
        self.from_imports = []
        self.node_factory = node_factory
        self.types = types
        self.type_vars = OrderedDict()
        self.file = file
        self.defined_var = None
        self.names_var = None
        if global_module and type_prefix != '__main__':
            self.add_builtin_vars()

    def add_builtin_vars(self) -> None:
        """
        Adds builtin variables that are defined in every module.
        """
        file_var = PythonGlobalVar('__file__', None,
                                   self.global_module.classes[STRING_TYPE], self)
        self.global_vars['__file__'] = file_var
        name_var = PythonGlobalVar('__name__', None,
                                   self.global_module.classes[STRING_TYPE], self)
        self.global_vars['__name__'] = name_var


    def process(self, translator: 'Translator') -> None:
        self.sil_name = self.get_fresh_name('module')
        defined_var_name = self.get_fresh_name('module_defined')
        self.defined_var = defined_var_name
        names_var_name = self.get_fresh_name('module_names')
        self.names_var = names_var_name

        for name, cls in self.classes.items():
            if name == cls.name:
                # if this is not a type alias
                cls.process(self.get_fresh_name(name), translator)
        for name, function in self.functions.items():
            function.process(self.get_fresh_name(name), translator)
        for name, method in self.methods.items():
            method.process(self.get_fresh_name(name), translator)
        for name, predicate in self.predicates.items():
            predicate.process(self.get_fresh_name(name), translator)
        for name, var in self.global_vars.items():
            var.process(self.get_fresh_name(name), translator)
        for name, operation in self.io_operations.items():
            operation.process(self.get_fresh_name(name), translator, self)

    @property
    def scope_prefix(self) -> List[str]:
        return []

    def get_func_or_method(self, name: str) -> 'PythonMethod':
        for module in [self] + self.from_imports + [self.global_module]:
            if not isinstance(module, PythonModule):
                # It's a ModuleView, we take the actual module wrapped by it.
                module = module.module
            if name in module.functions:
                return module.functions[name]
            elif name in module.methods:
                return module.methods[name]

    def get_type(self, prefixes: List[str], name: str,
                 previous: Tuple['PythonModule', ...] = ()) -> Tuple[str,
                                                                     Dict[Tuple[int, int],
                                                                          str]]:
        """
        Returns the main type and the alternative types of the element
        identified by this name found under this prefix in the current module
        (or imported ones).
        E.g., the type of local variable 'a' from method 'm' in class 'C'
        will be returned for the input (['C', 'm'], 'a').
        """
        if self in previous:
            return None, None
        actual_prefix = self.type_prefix.split('.') if self.type_prefix else []
        actual_prefix.extend(prefixes)
        local_type, local_alts = self.types.get_type(actual_prefix, name)
        if local_type is not None:
            return local_type, local_alts
        for module in self.from_imports:
            module_result = module.get_type(prefixes, name, previous + (self,))
            if module_result != (None, None):
                return module_result
        return None, None

    def get_func_type(self, path: List[str]):
        """
        Returns the type of the function identified by the given path in the
        current module (including imported other modules). It is assumed that
        the path points to a function, and the returned type will be the return
        type of that function, i.e., generally not a function type.
        """
        actual_prefix = self.type_prefix.split('.') if self.type_prefix else []
        actual_prefix.extend(path)
        local_result = self.types.get_func_type(actual_prefix)
        if local_result is not None:
            return local_result
        for module in self.from_imports:
            module_result = module.get_func_type(path)
            if module_result is not None:
                return module_result
        return None

    def get_included_modules(self, exclude: Set['PythonModule'] = (),
                             include_global: bool = True) -> List['PythonModule']:
        """
        Returns a list of modules whose contents are (transitively) available in the this
        module, optionally including the global module, but excluding the modules in the
        given set (to prevent infinite recursion in case of cyclic imports).
        """
        result = [self]
        for p in self.from_imports:
            result.extend(p.get_included_modules(exclude + (self,),
                                                 include_global=False))
        if include_global:
            result.append(self.global_module)
        return result

    def get_contents(self, only_top: bool) -> Dict:
        """
        Returns the elements that can be accessed from this container (to be
        used by get_target). If 'only_top' is true, returns only top level
        elements that can be accessed without a receiver.
        """
        dicts = [self.classes,  self.functions, self.global_vars, self.methods,
                 self.predicates, self.io_operations, self.namespaces]
        return CombinedDict([], dicts)


class PythonNode:
    def __init__(self, name: str, node=None):
        self.node = node
        self.name = name
        self.sil_name = None


class PythonType(metaclass=ABCMeta):
    """
    Abstract superclass of all kinds python types.
    """

    def try_box(self) -> 'PythonType':
        """
        If this class represents a primitive type, returns the boxed version,
        otherwise just return the type itself.
        """
        # By default, just return self. Subclasses can override.
        return self

    def try_unbox(self) -> 'PythonType':
        """
        If this class represents a boxed version of a primitive type, returns
        the primitive version, otherwise just returns the type itself.
        """
        # By default, just return self. Subclasses can override.
        return self


class SilverType(PythonType):
    """
    Wrapper around a Silver type. Only used for creating local variables with types not
    present in Python.
    """
    def __init__(self, type, module):
        self.type = type
        self.module = module


class TypeVar(PythonType, ContainerInterface):
    """
    Represents a type variable.
    """

    def __init__(self, name: str, bound: PythonType,
                 module: PythonModule, target_type: PythonType = None,
                 target_node: ast.AST = None,
                 index: int = None):
        """
        For class parameters, supply target_type, target_node and index. For
        method parameters, supply the defining node as target_node and later
        set self.type_expr.
        """
        self.name = name
        self.target_type = target_type
        self.target_node = target_node
        self.index = index
        self.bound = bound
        self.type_expr = None
        self.module = module

    def get_contents(self, only_top: bool) -> Dict:
        return self.bound.get_contents(only_top)

    @property
    def python_class(self) -> 'PythonClass':
        return self.bound


class PythonClass(PythonType, PythonNode, PythonScope, ContainerInterface):
    """
    Represents a class in the Python program.
    """

    def __init__(self, name: str, superscope: PythonScope,
                 node_factory: 'ProgramNodeFactory', node: ast.AST = None,
                 superclass: 'PythonClass' = None, interface=False):
        """
        :param superscope: The scope, usually module, this belongs to.
        :param interface: True iff the class implementation is provided in
        native Silver.
        """
        PythonNode.__init__(self, name, node)
        PythonScope.__init__(self, None, superscope)
        self.direct_subclasses = []
        self.node_factory = node_factory
        self.superclass = superclass
        if superclass:
            superclass.direct_subclasses.append(self)
        self.functions = OrderedDict()
        self.methods = OrderedDict()
        self.static_methods = OrderedDict()
        self.predicates = OrderedDict()
        self.fields = OrderedDict()
        self.static_fields = OrderedDict()
        self.type = None  # infer, domain type
        self.interface = interface
        self.defined = False
        self._has_classmethod = False
        self.type_vars = OrderedDict()
        self.definition_deps = set()
        self.is_adt = name == 'ADT' # This flag is set when the class is
        # defining an algebraic data type or one of its constructors.
        # This flag is set transitively across subclasses.

    @property
    def is_defining_adt(self) -> bool:
        """
        Returns true if the class is defining an algebraic data type. This class
        defines the name of the ADT and should be directly inherited by the
        classes defining the ADT's constructors.
        """
        assert self.is_adt
        if self.superclass:
            return self.superclass.name == 'ADT'
        return False

    @property
    def adt_def(self) -> 'PythonClass':
        """
        Returns the class that defines the ADT.
        """
        assert self.is_adt
        if self.is_defining_adt:
            return self
        else:
            return self.superclass.adt_def

    @property
    def adt_domain_name(self) -> str:
        """
        Returns the domain name where the ADT is defined.
        """
        assert self.is_adt
        return self.fresh(self.adt_def.name)

    @property
    def adt_prefix(self) -> str:
        """
        Returns the prefix of the domain name where the ADT is defined.
        Prefixes are used in defining domain functions and axioms.
        """
        assert self.is_adt
        return self.adt_def.name + '_'

    @property
    def all_subclasses(self) -> List['PythonClass']:
        """
        Returns all direct or indirect subclasses of this class.
        """
        res = [self]
        for sub in self.direct_subclasses:
            res.extend(sub.all_subclasses)
        return res

    @property
    def call_deps(self):
        """
        Returns the dependencies which need to be defined when calling this class, i.e.,
        its constructor.
        """
        constructor = self.get_method('__init__')
        if constructor:
            return constructor.call_deps
        return set()

    @property
    def all_methods(self) -> Set[str]:
        result = set()
        if self.superclass:
            result |= self.superclass.all_methods
        result |= set(self.methods.keys())
        return result

    @property
    def all_static_fields(self) -> Set[str]:
        result = set()
        if self.superclass:
            result |= self.superclass.all_static_fields
        result |= set(self.static_fields)
        return result

    def fresh(self, name: str) -> str:
        assert self.is_adt
        if not hasattr(self.adt_def, '_fresh'):
            self.adt_def._fresh = SingletonFreshName(self.module)
        return self.adt_def._fresh(name)

    def types_match(self, type_a: 'PythonType', type_b: 'PythonType') -> bool:
        """
        Checks if types are either the same or, if one of them is an union, checks
        if the other type belongs to the set of types in the union. Finally, if
        both types are unions, this function returns true if they have elements
        in common.
        """
        if not isinstance(type_a, UnionType) and isinstance(type_b, UnionType):
            return type_a in type_b.get_types()
        elif not isinstance(type_b, UnionType) and isinstance(type_a, UnionType):
            return type_b in type_a.get_types()
        elif isinstance(type_a, UnionType) and isinstance(type_b, UnionType):
            return len(type_a.get_types() & type_b.get_types()) > 0
        else:
            return type_a == type_b

    def add_field(self, name: str, node: ast.AST,
                  type: 'PythonType') -> 'PythonField':
        """
        Adds a field with the given name and type if it doesn't exist yet in
        this class.
        """
        if name in self.fields:
            field = self.fields[name]
            assert self.types_match(field.type.try_box(), type.try_box())
        elif name in self.static_fields:
            field = self.static_fields[name]
        else:
            field = self.node_factory.create_python_field(name, node,
                                                          type, self)
            self.fields[name] = field
        return field

    def get_field(self, name: str) -> Optional['PythonField']:
        """
        Returns the field with the given name in this class or a superclass.
        """
        if name in self.fields:
            return self.fields[name]
        elif self.superclass is not None:
            return self.superclass.get_field(name)
        else:
            return None

    def get_static_field(self, name: str) -> Optional['PythonVar']:
        """
        Returns the static field with the given name in this class or a
        superclass.
        """
        if name in self.static_fields:
            return self.static_fields[name]
        elif self.superclass is not None:
            return self.superclass.get_static_field(name)
        else:
            return None

    def get_method(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the method with the given name in this class or a superclass.
        """
        if name in self.methods:
            return self.methods[name]
        elif name in self.static_methods:
            return self.static_methods[name]
        elif self.superclass is not None:
            return self.superclass.get_method(name)
        else:
            return None

    def has_function(self, name: str) -> bool:
        """
        Check the function with the given name exists this class or
        superclass.
        """
        return name in self.functions or (self.superclass.has_function(name)
               if self.superclass is not None else False)

    def get_function(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the function with the given name in this class or a superclass.
        """
        if name in self.functions:
            return self.functions[name]
        elif self.superclass is not None:
            return self.superclass.get_function(name)
        else:
            return None

    def get_func_or_method(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the function or method with the given name in this class or a
        superclass.
        """
        if self.get_function(name) is not None:
            return self.get_function(name)
        else:
            return self.get_method(name)

    def get_predicate(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the predicate with the given name in this class or a superclass.
        """
        if name in self.predicates:
            return self.predicates[name]
        elif self.superclass is not None:
            return self.superclass.get_predicate(name)
        else:
            return None

    @property
    def all_fields(self) -> List['PythonField']:
        fields = []
        cls = self
        while cls is not None:
            for field in cls.python_class.fields.values():
                if isinstance(field, PythonField) and field.inherited is None:
                    fields.append(field)
            cls = cls.superclass
        return fields

    @property
    def all_sil_fields(self) -> List['silver.ast.Field']:
        """
        Returns a list of fields defined in the given class or its superclasses.
        """
        fields = []
        cls = self
        while cls is not None:
            for field in cls.python_class.fields.values():
                if isinstance(field, PythonField) and field.inherited is None:
                    fields.append(field.sil_field)
            cls = cls.superclass
        return fields

    @property
    def has_classmethod(self) -> bool:
        if self._has_classmethod:
            return True
        if self.superclass:
            return self.superclass.has_classmethod
        return False

    def process(self, sil_name: str, translator: 'Translator') -> None:
        """
        Creates fresh Silver names for all class members and initializes all
        of them.
        """
        self.sil_name = sil_name
        for name, function in self.functions.items():
            func_name = self.name + '_' + name
            function.process(self.get_fresh_name(func_name), translator)
        for name, method in self.methods.items():
            method_name = self.name + '_' + name
            method.process(self.get_fresh_name(method_name), translator)
        for name, method in self.static_methods.items():
            method_name = self.name + '_' + name
            method.process(self.get_fresh_name(method_name), translator)
        for name, predicate in self.predicates.items():
            pred_name = self.name + '_' + name
            predicate.process(self.get_fresh_name(pred_name), translator)
        for name, field in self.fields.items():
            if isinstance(field, PythonMethod):
                # This is a property.
                getter_name = self.name + '_' + name + '_getter'
                setter_name = self.name + '_' + name + '_setter'
                field.process(self.get_fresh_name(getter_name), translator)
                if field.setter:
                    field.setter.process(self.get_fresh_name(setter_name), translator)
            else:
                field_name = self.name + '_' + name
                field.process(self.get_fresh_name(field_name))
        for name, field in self.static_fields.items():
            field_name = self.name + '_' + name
            field.process(self.get_fresh_name(field_name), translator)
        if self.interface:
            all_methods = list(self.functions.values())
            all_methods.extend(self.methods.values())
            for method in all_methods:
                requires = set()
                for requirement in method.requires:
                    target = self.get_func_or_method(requirement)
                    if target:
                        requires.add(target.sil_name)
                    else:
                        requires.add(requirement)
                translator.set_required_names(method.sil_name, requires)

    def issubtype(self, cls: 'PythonClass') -> bool:
        if cls is self:
            return True
        if self.superclass is None:
            return False
        return self.superclass.issubtype(cls)

    def get_common_superclass(self, other: 'PythonClass') -> 'PythonClass':
        """
        Returns the common superclass of both classes. Raises an error if they
        don't have any, which should never happen.
        """
        if self.issubtype(other):
            return other
        if other.issubtype(self):
            return self
        if self.superclass:
            return self.superclass.get_common_superclass(other)
        elif other.superclass:
            return self.get_common_superclass(other.superclass)
        else:
            assert False, 'Internal error: Classes without common superclass.'

    def get_contents(self, only_top: bool) -> Dict:
        """
        Returns the elements that can be accessed from this container (to be
        used by get_target). If 'only_top' is true, returns only top level
        elements that can be accessed without a receiver.
        """
        dicts = [self.static_methods, self.static_fields]
        if not only_top:
            dicts.extend([self.functions, self.fields, self.methods,
                          self.predicates])
        return CombinedDict([], dicts)

    def try_box(self) -> 'PythonClass':
        """
        If this class represents a primitive type, returns the boxed version,
        otherwise just return the type itself.
        """
        if self.name in PRIMITIVES and self.name not in (PRIMITIVE_SEQ_TYPE + '_type',
                                                         PRIMITIVE_SET_TYPE + '_type',
                                                         CALLABLE_TYPE):
            boxed_name = self.name[len(PRIMITIVE_PREFIX):]
            if boxed_name == 'Set':
                boxed_name = PSET_TYPE
            if boxed_name == 'Multiset':
                boxed_name = PMSET_TYPE
            if boxed_name == 'Seq':
                boxed_name = PSEQ_TYPE
            return self.module.classes[boxed_name]
        return self

    def try_unbox(self) -> 'PythonClass':
        """
        If this class represents a boxed version of a primitive type, returns
        the primitive version, otherwise just returns the type itself.
        """
        if self.name in BOXED_PRIMITIVES:
            unboxed_name = PRIMITIVE_PREFIX + self.name
            return self.module.classes[unboxed_name]
        return self

    @property
    def python_class(self) -> 'PythonClass':
        return self


class GenericType(PythonType):
    """
    Represents a specific instantiation of a generic type, e.g. list[int].
    Provides access to the type arguments (in this case int) and otherwise
    behaves like the underlying PythonClass (in this case list).
    """

    def __init__(self, cls: PythonClass,
                 args: List[PythonType]) -> None:
        self.name = cls.name
        self.module = cls.module
        self.cls = cls
        self.type_args = args
        self.exact_length = True

    @property
    def python_class(self) -> PythonClass:
        return self.cls

    @property
    def sil_name(self) -> str:
        return self.python_class.sil_name

    def add_field(self, name: str, node: ast.AST,
                  type: 'PythonType') -> 'PythonField':
        return self.cls.add_field(name, node, type)

    @property
    def superclass(self) -> PythonClass:
        result = self.python_class.superclass
        if isinstance(result, GenericType):
            # Return a GenericType that has all type variables instantiated
            # based on the type arguments of this type.
            args = []
            for arg in result.type_args:
                if isinstance(arg, TypeVar):
                    args.append(self.type_args[arg.index])
                else:
                    args.append(arg)
            result = GenericType(result.cls, args)
        return result

    def get_field(self, name: str) -> Optional['PythonField']:
        """
        Returns the field with the given name in this class or a superclass.
        """
        return self.python_class.get_field(name)

    def get_method(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the method with the given name in this class or a superclass.
        """
        return self.python_class.get_method(name)

    def has_function(self, name: str) -> bool:
        """
        Check if all types of the generic type have a function with the given
        name.
        """
        if isinstance(self, UnionType):
            types_set = self.get_types() - {None}
            result = len(types_set) > 0
            for type in types_set:
                result = result and type.has_function(name)
            return result
        else:
            return self.cls.has_function(name)

    def get_function(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the function with the given name in this class or a superclass.
        """
        return self.python_class.get_function(name)

    def get_func_or_method(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the function or method with the given name in this class or a
        superclass.
        """
        return self.python_class.get_func_or_method(name)

    @property
    def all_methods(self) -> Set[str]:
        return self.python_class.all_methods

    @property
    def all_static_fields(self) -> Set[str]:
        return self.python_class.all_static_fields

    def get_predicate(self, name: str) -> Optional['PythonMethod']:
        """
        Returns the predicate with the given name in this class or a superclass.
        """
        return self.python_class.get_predicate(name)

    def issubtype(self, other: PythonType) -> bool:
        if isinstance(other, GenericType):
            if self.python_class.issubtype(other.cls):
                if self.type_args == other.type_args:
                    return True
            return False
        return self.python_class.issubtype(other)

    def get_contents(self, only_top: bool) -> Dict:
        return self.python_class.get_contents(only_top)

    def __hash__(self) -> int:
        return hash(self.python_class)

    def __eq__(self, other) -> bool:
        if not isinstance(other, GenericType):
            return False
        if self.cls != other.cls or len(self.type_args) != len(other.type_args):
            return False
        for my_arg, other_arg in zip(self.type_args, other.type_args):
            if my_arg != other_arg:
                return False
        return True


class UnionType(GenericType):
    """
    A special case of a generic type for union types. Behaves like any generic
    type named 'Union' with the given type arguments in most scenarios, but
    if you look up methods/functions/..., it will give you those offered by
    the common superclass of all its arguments (which should always exist and
    be 'object' if there is no other connection).
    In the special case of an optional type, it will just give you all the
    members of the non-None option.
    """
    def __init__(self, args: List[PythonType]) -> None:
        self.name = 'Union'
        self._cls = None
        self.module = args[0].module
        self.type_args = args
        self.exact_length = True

    @property
    def cls(self):
        if self._cls is None:
            # Get common supertype of all types in the union.
            args = self.type_args
            cls = args[0]
            if isinstance(cls, GenericType):
                cls = cls.cls
            for type_option in args[1:]:
                if type_option:
                    if isinstance(type_option, GenericType):
                        type_option = type_option.cls
                    cls = cls.get_common_superclass(type_option)
            self._cls = cls
        return self._cls

    @property
    def python_class(self) -> PythonClass:
        return self.cls.python_class

    def get_types(self) -> Set[PythonClass]:
        """
        Returns a flattened set of types contained in Union
        """
        result = set()
        for type in self.type_args:
            if not isinstance(type, UnionType):
                if type is None:
                    result.add(None)
                else:
                    result.add(type.python_class)
            else:
                result |= type.get_types()
        return result


class OptionalType(UnionType):
    """
    A special case of a union type for optional types, i.e., unions of some type
    and NoneType. Will behave like a normal union type. If you look up methods
    etc., it will behave like its type argument (e.g., Optional[C] would behave
    like C).
    """
    def __init__(self, typ: PythonType) -> None:
        super().__init__([typ])
        self.type_args = [None, typ]
        self.optional_type = typ

    @property
    def cls(self):
        return self.optional_type

    def get_types(self) -> Set[PythonClass]:
        """
        Returns a flattened set of types contained in Optional
        """
        if isinstance(self.optional_type, UnionType):
            return {None} | self.optional_type.get_types()
        else:
            return {None} | {self.optional_type}


class MethodType(Enum):
    normal = 0
    static_method = 1
    class_method = 2


def add_all_call_deps(call_deps: Set[Tuple[ast.AST, PythonNode, PythonModule]],
                      res: Set[Tuple[ast.AST, PythonNode, PythonModule]],
                      prefix: Tuple[PythonNode, ...]=()) -> None:
    """
    Adds all dependencies represented by call_deps to the given set.
    The set will contain tuples of length at least 3, where the first element is
    the Python AST node representing the access, the second the PythonNode accessed,
    the third the PythonModule in which the first name needs to be defined.
    All further elements are PythonNodes which represent conditions, i.e., the name
    needs to be defined in the module IF all the conditional PythonNodes have been
    defined in their respective modules.
    """
    for dep in call_deps:
        if dep not in res:
            c_prefix = prefix
            if len(dep) > 3:
                # this is a conditional dependency; its dependencies are to be in-
                # cluded under this condition.
                c_prefix = prefix + dep[3:]
            else:
                # this is a direct dependency, add it to the result
                res.add(dep + c_prefix)

            if hasattr(dep[1], 'add_all_call_deps'):
                dep[1].add_all_call_deps(res, c_prefix)


class PythonMethod(PythonNode, PythonScope, ContainerInterface, PythonStatementContainer):
    """
    Represents a Python function which may be pure or impure, belong
    to a class or not
    """

    def __init__(self, name: str, node: ast.AST, cls: PythonClass,
                 superscope: PythonScope,
                 pure: bool, contract_only: bool,
                 node_factory: 'ProgramNodeFactory',
                 interface: bool = False,
                 interface_dict: Dict[str, Any] = None,
                 method_type: MethodType = MethodType.normal):
        """
        :param cls: Class this method belongs to, if any.
        :param superscope: The scope (class or module) this method belongs to
        :param pure: True iff ir's a pure function, not an impure method
        :param contract_only: True iff we're not generating the method's
        implementation, just its contract
        :param interface: True iff the method implementation is provided in
        native Silver.
        """
        PythonNode.__init__(self, name, node)
        PythonScope.__init__(self, None, superscope)
        PythonStatementContainer.__init__(self)
        if cls is not None:
            if not isinstance(cls, PythonClass):
                raise Exception(cls)
        self.cls = cls
        self.overrides = None  # infer
        self._locals = OrderedDict()  # direct
        self.globals = set()
        self._args = OrderedDict()  # direct
        self._special_vars = OrderedDict()  # direct
        self._io_existential_vars = OrderedDict()
        self._nargs = -1  # direct
        self.var_arg = None   # direct
        self.kw_arg = None  # direct
        self.type = None  # infer
        self.generic_type = -1
        self.result = None  # infer
        self.error_var = None  # infer
        self.declared_exceptions = OrderedDict()  # direct
        self.pure = pure
        self.predicate = False
        self.all_low = False
        self.preserves_low = False
        self.contract_only = contract_only
        self.interface = interface
        self.interface_name = None  # Name to be used in error messages, if different from
                                    # the "actual" internal name. To be used for interface
                                    # methods.
        self.interface_dict = interface_dict
        self.node_factory = node_factory
        self.method_type = method_type
        self.obligation_info = None
        self.requires = []
        self.type_vars = OrderedDict()
        self.setter = None
        self.func_constant = None
        self.threading_id = None
        self.definition_deps = set()
        self.call_deps = set()

    def add_all_call_deps(self, res: Set[Tuple[ast.AST, PythonNode, PythonModule]],
                          prefix: Tuple[PythonNode, ...]=()) -> None:
        """
        Adds all dependencies needed when this method is called to the given set.
        """
        add_all_call_deps(self.call_deps, res, prefix)

    def process(self, sil_name: str, translator: 'Translator') -> None:
        """
        Creates fresh Silver names for all parameters and initializes them,
        same for local variables. Also sets the method type and
        checks if this method overrides one from a superclass,
        """
        self.sil_name = sil_name
        self.threading_id = self.superscope.get_fresh_name(self.name + "_threading")
        if self.pure:
            self.func_constant = self.superscope.get_fresh_name(self.name)
        for name, arg in self.args.items():
            arg.process(self.get_fresh_name(name), translator)
        if self.var_arg:
            self.var_arg.process(self.get_fresh_name(self.var_arg.name),
                                 translator)
        if self.kw_arg:
            self.kw_arg.process(self.get_fresh_name(self.kw_arg.name),
                                translator)
        self.obligation_info = translator.create_obligation_info(self)
        if self.interface:
            requires = set()
            for requirement in self.requires:
                requires.add(requirement)
            translator.set_required_names(self.sil_name, requires)
            return
        func_type = self.module.types.get_func_type(self.scope_prefix)
        if self.type is not None:
            self.result = self.node_factory.create_python_var(RESULT_NAME, None,
                                                              self.type)
            self.result.process(RESULT_NAME, translator)
        if self.cls is not None and self.cls.superclass is not None:
            try:
                # Could be overridden by anything, so we have to check if there's
                # anything with the same name.
                self.overrides = self.cls.superclass.get_contents(False)[self.name]
            except KeyError:
                pass
        for local in self.locals:
            self.locals[local].process(self.get_fresh_name(local), translator)
        for name in self.special_vars:
            self.special_vars[name].process(self.get_fresh_name(name),
                                            translator)
        for try_block in self.try_blocks:
            try_block.process(translator)

    @property
    def nargs(self) -> int:
        if self._nargs == -1:
            return len(self.args)
        else:
            return self._nargs

    @nargs.setter
    def nargs(self, nargs: int) -> None:
        self._nargs = nargs

    @property
    def args(self) -> OrderedDict:
        # TODO(shitz): Should make this return an immutable copy.
        return self._args

    def add_arg(self, name: str, arg: 'PythonVar'):
        self._args[name] = arg

    @property
    def locals(self) -> OrderedDict:
        # TODO(shitz): Should make this return an immutable copy.
        return self._locals

    def add_local(self, name: str, local: 'PythonVar'):
        self._locals[name] = local

    @property
    def special_args(self) -> Dict:
        result = {}
        if self.kw_arg:
            result[self.kw_arg.name] = self.kw_arg
        if self.var_arg:
            result[self.var_arg.name] = self.var_arg
        return result

    @property
    def special_vars(self) -> OrderedDict:
        return self._special_vars

    @property
    def io_existential_vars(self) -> OrderedDict:
        """
        IO existential variables are variables defined by using
        ``IOExists`` construct.
        """
        return self._io_existential_vars

    def get_variable(self, name: str) -> Optional['PythonVar']:
        """
        Returns the variable (local variable or method parameter) with the
        given name.
        """
        if name in self.locals:
            return self.locals[name]
        elif name in self.args:
            return self.args[name]
        elif name in self.special_vars:
            return self.special_vars[name]
        elif name in self.io_existential_vars:
            return self.io_existential_vars[name]
        elif self.var_arg and self.var_arg.name == name:
            return self.var_arg
        elif self.kw_arg and self.kw_arg.name == name:
            return self.kw_arg
        else:
            return self.module.global_vars.get(name)

    def create_variable(self, name: str, cls: PythonClass,
                        translator: 'Translator',
                        local: bool=True) -> 'PythonVar':
        """
        Creates a new local variable with the given name and type and performs
        all necessary processing/initialization
        """
        sil_name = self.get_fresh_name(name)
        result = self.node_factory.create_python_var(name, None, cls)
        result.process(sil_name, translator)
        if local:
            self.add_local(sil_name, result)
        return result

    def get_locals(self) -> List['PythonVar']:
        """
        Returns all method locals as a list of PythonVars.
        """
        return list(self.locals.values())

    def get_args(self) -> List['PythonVar']:
        """
        Returns all method args as a list of PythonVars.
        """
        return list(self.args.values())

    def get_results(self) -> List['PythonVar']:
        """
        Returns all results as a list of PythonVars.
        """
        if self.result:
            return [self.result]
        else:
            return []

    def get_contents(self, only_top: bool) -> Dict:
        """
        Returns the elements that can be accessed from this container (to be
        used by get_target). If 'only_top' is true, returns only top level
        elements that can be accessed without a receiver.
        """
        dicts = [self.args,  self.special_args, self.locals, self.special_vars]
        return CombinedDict([], dicts)




class PythonIOOperation(PythonNode, PythonScope, ContainerInterface):
    """
    Represents an IO operation which may be basic or not.

    +   ``preset`` – a set of input places.
    +   ``postset`` – a set of output places.
    +   ``inputs`` – inputs of IO operation (excluding preset).
    +   ``outputs`` – outputs of IO operation (excluding postset).
    """
    def __init__(
            self,
            name: str,
            node: ast.AST,
            superscope: PythonScope,
            node_factory: 'ProgramNodeFactory',
            ):
        PythonNode.__init__(self, name, node)
        PythonScope.__init__(self, None, superscope)
        self._preset = []
        self._postset = []
        self._inputs = []
        self._outputs = []
        self._terminates = None
        self._termination_measure = None
        self._body = None
        self._io_existentials = None
        self.func_args = []
        self.definition_deps = set()
        self.call_deps = set()

    def add_all_call_deps(self, res: Set[Tuple[ast.AST, PythonNode, PythonModule]],
                          prefix: Tuple[PythonNode, ...]=()) -> None:
        add_all_call_deps(self.call_deps, res, prefix)

    @property
    def is_builtin(self) -> bool:
        return self.name in BUILTIN_IO_OPERATIONS

    def _process_var_list(self, var_list: List['PythonVar'],
                          translator: 'Translator') -> None:
        """
        Creates fresh Silver names for all variables in ``var_list``.
        """
        for var in var_list:
            var.process(self.get_fresh_name(var.name), translator)

    def process(self, sil_name: str, translator: 'Translator',
                module: PythonModule) -> None:
        """
        Creates fresh Silver names for preset, postset, inputs and
        outputs. Also, sets the ``sil_name``.
        """
        if self._body is not None:
            body_checker = IOOperationBodyChecker(
                self._body,
                self.get_results(),
                self._io_existentials,
                module,
                translator)
            body_checker.check()
        self.sil_name = sil_name
        self._process_var_list(self._preset, translator)
        self._process_var_list(self._postset, translator)
        self._process_var_list(self._inputs, translator)
        self._process_var_list(self._outputs, translator)

    def set_preset(self, preset: List['PythonVar']) -> None:
        assert len(preset) == 1 or self.is_builtin
        self._preset = preset

    def set_postset(self, postset: List['PythonVar']) -> None:
        assert len(postset) == 1 or self.is_builtin
        self._postset = postset

    def set_inputs(self, inputs: List['PythonVar']) -> None:
        assert isinstance(inputs, list)
        self._inputs = inputs

    def is_input(self, name) -> bool:
        for input in self._inputs:
            if input.name == name:
                return True
        else:
            return False

    def set_outputs(self, outputs: List['PythonVar']) -> None:
        assert isinstance(outputs, list)
        self._outputs = outputs

    def set_terminates(self, expression: ast.AST) -> bool:
        """
        Set the property if it is not already set and return ``True``.
        """
        if self._terminates is None:
            self._terminates = expression
            return True
        else:
            return False

    def get_terminates(self) -> ast.AST:
        """
        Get the ``Terminates`` property.

        If it was not set, it sets a default ``False``.
        """
        if self._terminates is None:
            self._terminates = ast.NameConstant(False)
            self._terminates.lineno = self.node.lineno
            self._terminates.col_offset = self.node.col_offset
        return self._terminates

    def set_termination_measure(self, expression: ast.AST) -> bool:
        """
        Set the property if it is not already set and return ``True``.
        """
        if self._termination_measure is None:
            self._termination_measure = expression
            return True
        else:
            return False

    def get_termination_measure(self) -> ast.AST:
        """
        Get the ``TerminationMeasure`` property.

        If it was not set, it sets a default ``1``.
        """
        if self._termination_measure is None:
            self._termination_measure = ast.Num(1)
            self._termination_measure.lineno = self.node.lineno
            self._termination_measure.col_offset = self.node.col_offset
        return self._termination_measure

    def is_basic(self) -> bool:
        """
        Is this IO operation basic?
        """
        return self._body is None

    def set_body(self, expression: ast.AST) -> bool:
        """
        Set the body if it is not already set and return ``True``.
        """
        if self._body is None:
            self._body = expression
            return True
        else:
            return False

    def get_body(self) -> ast.AST:
        """
        Return IO operation body.
        """
        assert self._body is not None
        return self._body

    def set_io_existentials(
            self,
            io_existentials: List['PythonVarCreator']) -> bool:
        """
        Set io existentials list if not already set and return
        ``True``.
        """
        if self._io_existentials is None:
            self._io_existentials = io_existentials
            return True
        else:
            return False

    def get_io_existentials(self) -> List['PythonVarCreator']:
        """
        Get IO existentials.
        """
        assert self._io_existentials is not None
        return self._io_existentials

    def get_parameters(self) -> List['PythonVar']:
        """
        Return a list of parameters that uniquely identify IO operation
        instance.
        """
        return self._preset + self._inputs

    def get_results(self) -> List['PythonVar']:
        """
        Return a list of results for this IO operation.
        """
        return self._outputs + self._postset

    def get_variable(self, name: str) -> Optional['PythonVar']:
        """
        Returns the variable (existential variable or parameter) with
        the given name.
        """
        for var in self._preset:
            if var.name == name:
                return var
        for var in self._inputs:
            if var.name == name:
                return var
        if self._io_existentials:
            for var in self._io_existentials:
                if var.name == name:
                    return var.create_io_existential_variable_instance()
        return None

    def get_contents(self, only_top: bool) -> Dict:
        """
        Returns the elements that can be accessed from this container (to be
        used by get_target). If 'only_top' is true, returns only top level
        elements that can be accessed without a receiver.
        """
        return IOOperationContentDict(self)


class PythonExceptionHandler(PythonNode):
    """
    Represents an except-block belonging to a try-block.
    """

    def __init__(self, node: ast.AST, exception_type: PythonClass,
                 try_block: 'PythonTryBlock', handler_name: str, body: ast.AST,
                 exception_name: str):
        """
        :param exception_type: The type of exception this handler catches
        :param try_block: The try-block this handler belongs to
        :param handler_name: Label that this handler will get in Silver
        :param exception_name: Variable name for the exception in the block
        """
        super().__init__(handler_name, node=node)
        self.try_block = try_block
        self.body = body
        self.exception = exception_type
        self.exception_name = exception_name


class PythonTryBlock(PythonNode):
    """
    Represents a try-block, which may include except-blocks, an else-block
    and/or a finally-block.
    """

    def __init__(self, node: ast.AST, try_name: str,
                 node_factory: 'ProgramNodeFactory',
                 method: PythonStatementContainer,
                 protected_region: ast.AST):
        """
        :param node_factory: Factory to create PythonVar objects.
        :param method: Method this block is in
        :param protected_region: Statements protected by the try
        """
        super().__init__(try_name, node=node)
        self.handlers = []
        self.else_block = None
        self.finally_block = None
        self.protected_region = protected_region
        self.finally_var = None
        self.error_var = None
        self.method = method
        self.node_factory = node_factory
        self.finally_name = None
        self.post_name = None
        self.with_item = None
        self.with_var = None
        self.handler_aliases = {}
        method.labels.append(try_name)

    def get_finally_var(self, translator: 'Translator') -> 'PythonVar':
        """
        Lazily creates and returns the variable in which we store the
        information how control flow should proceed after the execution
        of a finally block. We use a value of 0 to say normal flow, 1 to say
        we returned normally, i.e. jump to the end of the function asap, 2
        to say we returned exceptionally, i.e. jump to the end of the function
        moving through exception handlers.
        """
        if self.finally_var:
            return self.finally_var
        sil_name = self.method.get_fresh_name('try_finally')
        global_module = self.method.module.global_module
        int_type = global_module.classes[PRIMITIVE_INT_TYPE]
        result = self.node_factory.create_python_var(sil_name, None,
                                                     int_type)
        result.process(sil_name, translator)
        if isinstance(self.method, PythonMethod):
            self.method.locals[sil_name] = result
        self.finally_var = result
        return result

    def get_error_var(self, translator: 'Translator') -> 'PythonVar':
        """
        Lazily creates and returns the variable in which any exceptions thrown
        within the block will be stored.
        """
        if self.error_var:
            return self.error_var
        sil_name = self.method.get_fresh_name('error')
        exc_type = self.method.module.global_module.classes['Exception']
        result = self.node_factory.create_python_var(sil_name, None,
                                                     exc_type)
        result.process(sil_name, translator)
        if isinstance(self.method, PythonMethod):
            self.method.locals[sil_name] = result
        self.error_var = result
        return result

    def process(self, translator: 'Translator') -> None:
        self.get_error_var(translator)
        self.get_finally_var(translator)


class PythonVarBase(PythonNode):
    """
    Abstract class representing any variable in Python.
    """

    def __init__(self, name: str, node: ast.AST, type: PythonClass):
        super().__init__(name, node)
        self.type = type
        self.writes = []
        self.reads = []
        self.alt_types = {}
        self.default = None
        self.default_expr = None

    def process(self, sil_name: str, translator: 'Translator') -> None:
        """
        Sets ``sil_name``.
        """
        self.sil_name = sil_name

    def get_specific_type(self, node: ast.AST) -> PythonType:
        """
        Returns the type of this variable when referenced by the given node,
        i.e., checks if there is an alt_type for this node, otherwise returns
        the default type.
        """
        col = get_column(node)
        key = (node.lineno, col)
        if key in self.alt_types:
            return self.alt_types[key]
        else:
            return self.type


class PythonVar(PythonVarBase, abc.ABC):
    """
    Represents a variable in Python. Can be a local variable or a
    function parameter.
    """

    def __init__(self, name: str, node: ast.AST, type: PythonClass):
        super().__init__(name, node, type)
        self.decl = None
        self._ref = None
        self._translator = None
        self.value = None

    def process(self, sil_name: str, translator: 'Translator') -> None:
        """
        Creates a Silver variable declaration and reference representing
        this Python variable.
        """
        super().process(sil_name, translator)
        self._translator = translator
        module = self.type.module
        self.decl = translator.translate_pythonvar_decl(self, module)
        self._ref = translator.translate_pythonvar_ref(self, module, None, None)

    def ref(self, node: ast.AST=None,
            ctx: 'Context'=None) -> 'silver.ast.LocalVarRef':
        """
        Creates a reference to this variable. If no arguments are supplied,
        the reference will have no position. Otherwise, it will have the
        position of the given node in the given context.
        """
        if not node:
            return self._ref
        module = self.type.module
        return self._translator.translate_pythonvar_ref(self, module, node, ctx)


class PythonGlobalVar(PythonVarBase):
    """
    Represents a global variable in Python.
    """

    def __init__(self, name: str, node: ast.AST, type: PythonClass, module: PythonModule,
                 cls: PythonClass = None):
        super().__init__(name, node, type)
        self.module = module
        self.cls = cls
        self.overrides = None

    @property
    def is_final(self) -> bool:
        """
        A variable is final if it is written to only once (globally). Built-in module
        variables like __name__ are not considered final.
        """
        return len(self.writes) <= 1 and self.name not in MODULE_VARS

    def process(self, sil_name: str, translator: 'Translator') -> None:
        super().process(sil_name, translator)
        if self.cls is not None and self.cls.superclass is not None:
            try:
                self.overrides = self.cls.superclass.get_contents(False)[self.name]
                if not isinstance(self.overrides, PythonGlobalVar):
                    raise InvalidProgramException(self.node, 'invalid.override')
            except KeyError:
                pass


class PythonIOExistentialVar(PythonVarBase):
    """
    Represents an existential variable in Python. Existential variable
    is a variable created by using ``IOExists`` construct and it can be
    used only in contracts. Unlike normal variables, it is translated to
    getter functions.
    """

    def __init__(self, name: str, node: ast.AST, type: PythonClass):
        super().__init__(name, node, type)
        self._ref = None
        self._old_ref = None

    def is_defined(self) -> bool:
        """
        Returns true if main getter was already defined.
        """
        return not self._ref is None

    def ref(self, node: ast.AST = None, ctx: 'Context' = None) -> Expr:
        """
        Returns a Silver expression node that can be used to refer to
        this variable.
        """
        # TODO (Vytautas): Update to new API.
        assert not self._ref is None, self.name
        if ctx.obligation_context.is_translating_posts:
            assert not self._old_ref is None, self.name
            return self._old_ref
        else:
            return self._ref

    def set_ref(self, ref: Expr, old_ref: Optional[Expr]) -> None:
        """
        Sets a Silver expression node that can be used to refer to this
        variable.
        """
        # TODO (Vytautas): Update to new API.
        assert self._ref is None
        assert self._old_ref is None
        assert ref is not None
        self._ref = ref
        self._old_ref = old_ref


class PythonVarCreator:
    """
    A factory for Python variable. It is used instead of a concrete
    variable when the same Python variable can be translated into
    multiple silver variables. For example, when opening a non-basic IO
    operation.
    """

    def __init__(self, name: str, node: ast.AST,
                 type: PythonClass) -> None:
        self._name = name
        self._node = node
        self._type = type

        # Information needed to construct defining getter.
        self._defining_order = None     # type: Optional[int]
        self._defining_node = None
        self._defining_result = None

        # Defining getter.
        self._existential_ref = None

    @property
    def defining_order(self) -> int:
        """In what order existentials are defined?

        If the variable's defining order is ``x``, then its defining
        getter might have any existential variable with defining order
        ``y (y < x)`` as its argument.
        """
        assert self._defining_order is not None
        return self._defining_order

    @property
    def name(self) -> str:
        return self._name

    def set_defining_info(self, order: int, node: ast.Call,
                          result: PythonVar) -> None:
        """Store information needed to contstruct the defining getter."""
        assert self._defining_order is None
        assert self._defining_node is None
        assert self._defining_result is None

        self._defining_order = order
        self._defining_node = node
        self._defining_result = result

    def get_defining_info(self) -> Tuple[ast.Call, PythonVar]:
        """Retrieve information needed to contstruct the defining getter."""
        assert self._defining_order is not None
        assert self._defining_node is not None
        assert self._defining_result is not None

        return self._defining_node, self._defining_result

    def set_existential_ref(self, ref: Expr) -> None:
        """Set defining getter."""
        assert not self._existential_ref
        self._existential_ref = ref

    def create_variable_instance(self) -> PythonVar:
        """Create a normal variable.

        Normal variables are used when translating ``Open`` statement.
        """
        return PythonVar(self._name, self._node, self._type)

    def create_io_existential_variable_instance(
            self) -> PythonIOExistentialVar:
        """Create an existential variable.

        Existential variables are used when translating termination
        checks.
        """
        var = PythonIOExistentialVar(self._name, self._node, self._type)
        assert self._existential_ref
        var.set_ref(self._existential_ref, None)
        return var


class PythonField(PythonNode):
    """
    Represents a field of a Python class.
    """
    def __init__(self, name: str, node: ast.AST, type: PythonClass,
                 cls: PythonClass):
        """
        :param type: The type of the field.
        :param cls: The class this field belongs to
        """
        super().__init__(name, node)
        self.cls = cls
        self.inherited = None  # infer
        self.type = type
        self._sil_field = None
        self.reads = []  # direct
        self.writes = []  # direct

    @property
    def sil_field(self) -> 'silver.ast.Field':
        return self._sil_field

    @sil_field.setter
    def sil_field(self, field: 'silver.ast.Field'):
        self._set_sil_field(field)

    def _set_sil_field(self, field: 'silver.ast.Field'):
        self._sil_field = field

    def process(self, sil_name: str) -> None:
        """
        Checks if this field is inherited from a superclass.
        """
        self.sil_name = sil_name
        if not self.is_mangled():
            if self.cls.superclass is not None:
                self.inherited = self.cls.superclass.get_field(self.name)

    def is_mangled(self) -> bool:
        return self.name.startswith('__') and not self.name.endswith('__')

    @property
    def actual_field(self) -> 'PythonField':
        """
        If this field is inherited from a superclass, it will not actually be used in the
        translation; this function will return the field that is actually used.
        """
        result = self
        while result.inherited is not None:
            result = result.inherited
        return result


class ProgramNodeFactory:
    """
    Factory to create Python ProgramNodes.

    TODO: Add more interfaces for other types of containers if needed.
    """

    def create_python_var(
            self, name: str, node: ast.AST,
            type_: PythonClass) -> PythonVar:
        return PythonVar(name, node, type_)

    def create_python_global_var(
            self, name: str, node: ast.AST, type_: PythonClass,
            module: PythonModule) -> PythonGlobalVar:
        return PythonGlobalVar(name, node, type_, module)

    def create_static_field(
            self, name: str, node: ast.AST, type_: PythonClass, module: PythonModule,
            cls: PythonClass) -> PythonGlobalVar:
        return PythonGlobalVar(name, node, type_, module, cls)

    def create_python_io_existential_var(
            self, name: str, node: ast.AST,
            type_: PythonClass) -> PythonIOExistentialVar:
        return PythonIOExistentialVar(name, node, type_)

    def create_python_var_creator(
            self, name: str, node: ast.AST,
            type_: PythonClass) -> PythonIOExistentialVar:
        return PythonVarCreator(name, node, type_)

    def create_python_method(
            self, name: str, node: ast.AST, cls: PythonClass,
            superscope: PythonScope,
            pure: bool, contract_only: bool,
            container_factory: 'ProgramNodeFactory',
            interface: bool = False,
            interface_dict: Dict[str, Any] = None,
            method_type: MethodType = MethodType.normal) -> PythonMethod:
        return PythonMethod(name, node, cls, superscope, pure, contract_only,
                            container_factory, interface, interface_dict,
                            method_type)

    def create_python_io_operation(self, name: str, node: ast.AST,
                                   superscope: PythonScope,
                                   container_factory: 'ProgramNodeFactory',
                                   ) -> PythonIOOperation:
        return PythonIOOperation(name, node, superscope, container_factory)

    def create_python_field(self, name: str, node: ast.AST, type_: PythonClass,
                            cls: PythonClass) -> PythonField:
        return PythonField(name, node, type_, cls)

    def create_python_class(self, name: str, superscope: PythonScope,
                            node_factory: 'ProgramNodeFactory',
                            node: ast.AST = None,
                            superclass: PythonClass = None,
                            interface=False):
        return PythonClass(name, superscope, node_factory, node,
                           superclass, interface)

def toposort_classes(class_set: Set[PythonClass]) -> List[PythonClass]:
    """
    Topological sorting of classes in a set, ensuring that derived classes
    precede their base classes in the returned list
    """
    map = {}

    for type in class_set:
        map[type] = set(type.all_subclasses) & class_set

    return list(toposort_flatten(map, False))

def chain_if_stmts(guarded_blocks: List[Tuple[Expr, Stmt]],
                   viper, position, info, ctx) -> Stmt:
    """
    Receives a list of tuples each one containing a guard and a guarded
    block and produces an equivalent chain of if statements.
    """
    assert(guarded_blocks)
    guard, then_block = guarded_blocks[0]
    if len(guarded_blocks) == 1:
        else_block = viper.Seqn([], position, info) # Empty block
    else:
        else_block = chain_if_stmts(guarded_blocks[1:], viper, position, info, ctx)
    return viper.If(guard, then_block, else_block, position, info)

def chain_cond_exp(guarded_expr: List[Tuple[Expr, Expr]],
                   viper, position, info, ctx) -> Expr:
    """
    Receives a list of tuples each one containing a guard and a guarded
    expression and produces an equivalent chain of conditional expressions.
    """
    if len(guarded_expr) == 1:
        return guarded_expr[0][1]
    guard, then_exp = guarded_expr[0]
    if len(guarded_expr) == 2:
        _, else_exp = guarded_expr[1]
    else:
        else_exp = chain_cond_exp(guarded_expr[1:], viper, position, info, ctx)
    return viper.CondExp(guard, then_exp, else_exp, position, info)


"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Code for handling errors.

This includes:

1.  Converting Silver level errors to Nagini errors.
2.  Creating human readable error messages.
"""


from nagini_translation.lib.errors.manager import manager as error_manager
from nagini_translation.lib.errors.rules import Rules



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Code for constructing Silver Method call node with obligation stuff."""


import ast

from typing import List, Optional

from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonMethod,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.node_constructor import (
    StatementNodeConstructorBase,
)


class ObligationMethodCall:
    """Info for generating Silver ``MethodCall`` AST node."""

    def __init__(
            self, name: str, args: List[Expr], targets: List[Expr]) -> None:
        assert name
        assert all(args)
        assert all(targets)
        self.name = name
        self.original_args = args[:]
        self.args = args
        self.targets = targets

    def prepend_arg(self, arg: Expr) -> None:
        """Prepend ``args`` to the argument list."""
        assert arg
        self.args.insert(0, arg)

    def prepend_target(self, target: Expr) -> None:
        """Prepend ``target`` to the target list."""
        assert target
        self.targets.insert(0, target)


class ObligationMethodCallNodeConstructor(StatementNodeConstructorBase):
    """A class that creates a method call node with obligation stuff."""

    def __init__(
            self, obligation_method_call: ObligationMethodCall,
            position: Position, info: Info,
            translator: 'AbstractTranslator', ctx: Context,
            obligation_manager: ObligationManager,
            target_method: Optional[PythonMethod],
            target_node: Optional[ast.AST]) -> None:
        """Constructor.

        If ``target_node`` is ``None``, then:

        1.  It is assumed that we are translating a behavioural
            subtyping check.
        2.  ``target_method.node`` is used for computing default
            position.
        """
        super().__init__(
            translator, ctx, obligation_manager, position, info,
            target_node or target_method.node)
        self._obligation_method_call = obligation_method_call
        self._target_method = target_method
        self._target_node = target_node

    def construct_call(self) -> None:
        """Construct statements to perform a call."""
        self._add_additional_arguments()
        self._add_additional_targets()
        self._add_call()

    def _add_additional_arguments(self) -> None:
        """Add current thread and caller measure map arguments."""
        if obligation_config.disable_all:
            return
        if self._ctx.obligation_context.is_translating_loop():
            loop_info = self._ctx.obligation_context.current_loop_info
            residue_level = loop_info.residue_level
        else:
            residue_level = self._obligation_info.residue_level
        self._obligation_method_call.prepend_arg(
            residue_level.ref(self._target_node, self._ctx))
        if not obligation_config.disable_measures:
            self._obligation_method_call.prepend_arg(
                self._obligation_info.method_measure_map.get_var().ref(
                    self._target_node, self._ctx))
        self._obligation_method_call.prepend_arg(
            self._obligation_info.current_thread_var.ref(
                self._target_node, self._ctx))

    def _add_additional_targets(self) -> None:
        """Add current wait level dummy target."""
        if obligation_config.disable_waitlevel_check:
            return
        self._obligation_method_call.prepend_target(
            self._obligation_info.current_wait_level_target.ref())

    def _add_call(self) -> None:
        """Add the actual call node."""
        call = self._obligation_method_call
        statement = self._viper.MethodCall(
            call.name, call.args, call.targets,
            self._position, self._info)
        self._statements.append(statement)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.lib.constants import (
    CALLABLE_TYPE,
    PRIMITIVES,
)
from nagini_translation.lib.program_nodes import (
    PythonClass,
    PythonIOOperation,
    PythonMethod,
    PythonType,
    SilverType,
)
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.resolver import get_type as do_get_type
from nagini_translation.lib.typedefs import (
    Expr,
)
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.abstract import (
    Context,
    TranslatorConfig,
)
from nagini_translation.translators.common import CommonTranslator
from typing import Optional


class TypeTranslator(CommonTranslator):

    def __init__(self, config: TranslatorConfig, jvm: JVM, source_file: str,
                 type_info: TypeInfo, viper_ast: ViperAST) -> None:
        super().__init__(config, jvm, source_file, type_info, viper_ast)

    @property
    def builtins(self):
        return {'builtins.int': self.viper.Int,
                'builtins.bool': self.viper.Bool,
                'builtins.PSeq': self.viper.SeqType(self.viper.Ref),
                'builtins.PSet': self.viper.SetType(self.viper.Ref),
                'builtins.PMultiset': self.viper.MultisetType(self.viper.Ref),
                }

    def translate_type(self, cls: PythonClass,
                       ctx: Context) -> 'silver.ast.Type':
        """
        Translates the given type to the corresponding Viper type (Int, Ref, ..)
        """
        if isinstance(cls, SilverType):
            return cls.type
        elif cls.name == CALLABLE_TYPE:
            return self.viper.function_domain_type()
        elif cls.name in PRIMITIVES:
            cls = cls.try_box()
            return self.builtins['builtins.' + cls.name]
        elif cls.name == 'type':
            return self.type_factory.type_type()
        else:
            return self.viper.Ref

    def get_type(self, node: ast.AST, ctx: Context) -> Optional[PythonType]:
        """
        Returns the type of the expression represented by node as a PythonType,
        or None if the type is void.
        """
        container = ctx.actual_function if ctx.actual_function else ctx.module
        containers = [ctx]
        if ctx.current_class:
            containers.append(ctx.current_class)
        if isinstance(container, (PythonMethod, PythonIOOperation)):
            containers.append(container)
            containers.extend(container.module.get_included_modules())
        else:
            # Assume module
            containers.extend(container.get_included_modules())
        return do_get_type(node, containers, container)

    def type_check(self, lhs: Expr, type: PythonType,
                   position: 'silver.ast.Position',
                   ctx: Context, inhale_exhale: bool=True) -> Expr:
        """
        Returns a type check expression. This may return a simple isinstance
        for simple types, or include information about type arguments for
        generic types, or things like the lengths for tuples.
        """
        inhale_exhale = False
        if type is None:
            none_type = ctx.module.global_module.classes['NoneType']
            return self.type_factory.type_check(lhs, none_type, position, ctx)
        elif type.name == 'type':
            return self.viper.TrueLit(position, self.no_info(ctx))
        else:
            result = self.type_factory.type_check(lhs, type, position, ctx)
            return result



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Top level Silver constructs."""


from nagini_translation.lib.silver_nodes.base import Node
from nagini_translation.lib.typedefs import Info
from nagini_translation.lib.typedefs import Field as FieldType
from nagini_translation.lib.typedefs import Position
from nagini_translation.lib.typedefs import Predicate as PredicateType


class Predicate(Node):
    """Bodyless predicate with one ``Ref`` argument definition."""

    def __init__(self, name: str, var_name: str) -> None:
        self._name = name
        self._var_name = var_name

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> PredicateType:
        var = translator.viper.LocalVarDecl(
            self._var_name, translator.viper.Ref, position, info)
        return translator.viper.Predicate(
            self._name, [var], None, position, info)


class Field(Node):
    """Field definition."""

    def __init__(self, name: str, typ: 'Type') -> None:
        self._name = name
        self._type = typ

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> FieldType:
        typ = self._type.translate(translator)
        return translator.viper.Field(
            self._name, typ, position, info)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

# pragma pylint: disable=wrong-spelling-in-docstring

r"""``MustTerminate`` obligation implementation.

The reasons why we do not follow original paper [Obligations]_ directly
are:

1.  We want to reuse Viper constructs as much as possible, which means
    that we would like to avoid replacing a method call with an exhale
    inhale pair.
2.  We do not want to have an additional argument, which indicates if
    the caller has an obligation to terminate or not.
3.  We want to have a consistent encoding for loops and methods.

``MustTerminate`` obligation encoding implemented in Chalice2Silver
([C2SObligations]_) turned out to be unsound (see issue `84
<https://bitbucket.org/viperproject/chalice2silver/issues/84/>`_). The
problem is that the callee's promise to terminate is checked after the
method call. If callee's postcondition is ``False`` (a possible
postcondition for non-terminating method), the check trivially passes.
Nagini therefore uses a different encoding.

``MustTerminate`` in Contracts
==============================

Like in [C2SObligations]_, ``MustTerminate`` mentioned in method
preconditions and loop invariants are translated into inhale/exhale
pair:

+   Unlike in [C2SObligations]_, we do not need to guarantee that we
    have at most full permission to ``MustTerminate``. We therefore just
    inhale a full permission to ``MustTerminate`` each time.
+   We do not exhale ``MustTerminate`` obligation (unlike in
    [C2SObligations]_), we only check that measure is strictly positive.

Promise to Terminate
====================

As described in the previous section, we do not exhale
``MustTerminate``. We therefore need a different mechanism for detecting
if a method call / loop promised to terminate. The idea is to generate a
boolean expression that is ``True`` iff method call / loop promised to
terminate. That is if there is at least one ``MustTerminate`` whose
guarding condition and measure check evaluates to ``True``. More
precisely, we define termination condition as:

.. math::

    tcond := \lor \left\{%
        guard(o) \land measure_check(o) :
        o \in \text{MustTerminate obligations in contract}
    \right\}

We also define a version that ignores measures:

.. math::

    tcond_no_measure := \lor \left\{%
        guard(o) :
        o \in \text{MustTerminate obligations in contract}
    \right\}

.. note::

    In subsequent sections :math:`tcond` and :math:`tcond_no_measure`
    are used as a macros.

:math:`tcond` and :math:`tcond_no_measure` implementation is provided by
:py:class:`BaseObligationInfo.create_termination_check`.

Method Encoding
===============

At the end of each method precondition, we add a leak check guarded by
:math:`tcond`. The whole expression added to the precondition in the
optimized form:

.. math::

    tcond \lor
    (perm(cthread) == none \land \text{leak check for other obligation types})

:math:`perm(cthread) == none` is an optimized form of the leak check for
``MustTerminate`` obligation. Note that in our encoding like in
[C2SObligations]_, leak check does not include ``MustTerminate``
obligation type.

.. note::

    Unlike in [C2SObligations]_, we do not emit any additional code on
    the caller side.

.. note::

    For axiomatic methods (the ones that are defined in
    ``preamble.index``) we add additional precondition:

    1.  If method is marked as terminating, we add a check that caller
        termination measure is bigger than 1:

        .. math::

            measure_check(cthread, 1)

    2.  If method is not marked as terminating, we add a leak check:

        .. math::

            (perm(cthread) == none \land
                \text{leak check for other obligation types})

Loop Encoding
=============

1.  We save ``MustTerminate`` amount in a variable so that we can
    restore it after the loop (like in [C2SObligations]_). Otherwise,
    terminating loop in non-terminating method would generate
    termination obligations.
2.  We save loop's termination promise:
    :math:`termination_flag := tcond_no_measure`.
3.  Like in [C2SObligations]_, to distinguish if we are exhaling before
    the loop or after the loop, we use a boolean local variable
    ``exhale_before_loop_var``. Before loop it is set to ``True`` and at
    the end of the loop body – to ``False``.
4.  At the end of the loop invariant (similarly to [C2SObligations]_,
    but the check is not the same), we add a leak check that guarantees
    that either loop promised to terminate, or that context does not
    have any obligations:

    .. math::

        exhale_before_loop_var \Rightarrow
            \not{loop_condition} \lor
            termination_flag \lor
                (perm(cthread) == none \land
                    \text{leak check for other obligation types})

    .. note::

        This check is almost identical to the one for method encoding.

5.  At the end of the loop invariant (similarly to [C2SObligations]_),
    we add a leak check that guarantees that loop body does not leak
    obligations:

    .. math::

        !exhale_before_loop_var \Rightarrow (
            \text{leak check for other obligation types}
        )

    .. note::

        This leak check does not check if loop upholds its termination
        promise.

6.  At the end of loop body we add a check that loop upholds its promise
    to terminate:

    .. math::

        termination_flag \Rightarrow (tcond \lor \not{loop_condition})

.. rubric:: References

..  [C2SObligations]
    Verification of Finite Blocking in Chalice
    Robert Meier, SS 2015
    https://www.ethz.ch/content/dam/ethz/special-interest/infk/chair-program-method/pm/documents/Education/Theses/Robert%20_Meier_MA_report.pdf

..  [Obligations]
    Modular Verification of Finite Blocking in Non-terminating Programs
    P. Boström and P. Müller
    European Conference on Object-Oriented Programming (ECOOP), 2015.
    http://pm.inf.ethz.ch/publications/getpdf.php?bibname=Own&id=BostromMueller15.pdf
"""


import ast

from typing import Any, Dict, List, Optional

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonMethod,
    PythonVar,
)
from nagini_translation.lib.typedefs import (
    Predicate,
)
from nagini_translation.translators.common import CommonTranslator
from nagini_translation.translators.obligation.inexhale import (
    InexhaleObligationInstanceMixin,
    ObligationInhaleExhale,
)
from nagini_translation.translators.obligation.types.base import (
    ObligationInstance,
    Obligation,
)


_OBLIGATION_NAME = 'MustTerminate'
_PREDICATE_NAME = _OBLIGATION_NAME


def _create_predicate_access(cthread: PythonVar) -> sil.PredicateAccess:
    """Create a predicate access expression."""
    return sil.PredicateAccess(_PREDICATE_NAME, sil.RefVar(cthread))


class MustTerminateObligationInstance(
        InexhaleObligationInstanceMixin, ObligationInstance):
    """Class representing instance of ``MustTerminate`` obligation."""

    def __init__(
            self, obligation: 'MustTerminateObligation', node: ast.expr,
            measure: ast.expr, target: PythonVar) -> None:
        super().__init__(obligation, node)
        self._measure = measure
        self._target = target

    def _get_inexhale(self, ctx: Context) -> ObligationInhaleExhale:
        return ObligationInhaleExhale(
            _create_predicate_access(self._target),
            skip_exhale=True,
            skip_inhale=obligation_config.disable_termination_check)

    def get_obligation_bound(self, ctx: Context) -> sil.Statement:
        assert False    # MustTerminate is never fresh.

    def is_fresh(self) -> bool:
        return False    # MustTerminate is never fresh.

    def get_measure(self) -> sil.IntExpression:
        return sil.PythonIntExpression(self._measure)

    def get_target(self) -> sil.RefExpression:
        return sil.RefVar(self._target)


class MustTerminateObligation(Obligation):
    """Class representing ``MustTerminate`` obligation."""

    def __init__(self) -> None:
        super().__init__([_PREDICATE_NAME], [])

    def identifier(self) -> str:
        return _OBLIGATION_NAME

    def check_node(
            self, node: ast.Call,
            obligation_info: 'PythonMethodObligationInfo',
            method: PythonMethod) -> Optional[MustTerminateObligationInstance]:
        if (isinstance(node.func, ast.Name) and
                node.func.id == _OBLIGATION_NAME):
            measure = node.args[0]
            instance = MustTerminateObligationInstance(
                self, node, measure, obligation_info.current_thread_var)
            return instance
        elif (isinstance(node.func, ast.Name) and
              node.func.id == "TerminatesSif"):
              measure = node.args[1]
              instance = MustTerminateObligationInstance(
                  self, node, measure, obligation_info.current_thread_var)
              return instance
        else:
            return None

    def create_predicate_access(
            self, cthread: PythonVar) -> sil.PredicateAccess:
        """Create a predicate access expression."""
        return _create_predicate_access(cthread)

    def is_interface_method_terminating(
            self, interface_dict: Dict[str, Any]) -> bool:
        """Check if interface method is annotated as terminating or not."""
        return (_OBLIGATION_NAME in interface_dict and
                interface_dict[_OBLIGATION_NAME])

    def generate_axiomatized_preconditions(
            self, obligation_info: 'PythonMethodObligationInfo',
            interface_dict: Dict[str, Any]) -> List[sil.BoolExpression]:
        """Add ``MustTerminate(1)`` to axiomatic method precondition."""
        if self.is_interface_method_terminating(interface_dict):
            cthread = sil.RefVar(obligation_info.current_thread_var)
            check = obligation_info.caller_measure_map.check(
                cthread, sil.RawIntExpression(1))
            return [check]
        else:
            return []

    def create_leak_check(self, var_name: str) -> List[sil.BoolExpression]:
        return [self._create_predicate_for_perm(_PREDICATE_NAME, var_name)]

    def create_predicates(
            self, translator: CommonTranslator) -> List[Predicate]:
        if obligation_config.disable_termination_check:
            return []
        else:
            return super().create_predicates(translator)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import argparse
import zmq

from nagini_translation.lib.constants import DEFAULT_CLIENT_SOCKET


context = zmq.Context()

socket = context.socket(zmq.REQ)
socket.connect(DEFAULT_CLIENT_SOCKET)

parser = argparse.ArgumentParser()
parser.add_argument(
        'python_file',
        help='Python file to verify')

args = parser.parse_args()

socket.send_string(args.python_file)
response = socket.recv_string()

print(response)


"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import copy

from typing import Dict, List, Set, Tuple


class CombinedDict:
    """
    A combined view on the given list of dicts, that adds the possibility to
    filter and rename some keys. E.g. if names contains ('k1', 'k2') then
    getting 'k1' will try to retrieve the value for key 'k2' from (one of) the
    given dicts. Unless 'names' is empty, keys that do not occur in 'names'
    will be reported as not contained by this view.
    """
    def __init__(self, names: List[Tuple[str, str]], dicts: List[Dict]):
        self.dicts = dicts
        self.names = {}
        for name, as_name in names:
            new_name = as_name if as_name else name
            self.names[new_name] = name

    def __getitem__(self, item):
        key = item
        if self.names:
            if not key in self.names:
                raise KeyError(item)
            key = self.names[key]
        for d in self.dicts:
            if key in d:
                return d[key]
        raise KeyError(item)

    def __contains__(self, item):
        return self.contains(item, None)

    def contains(self, item, caller_module):
        """
        Checks if the given item is contained anywhere in this view, excluding anything
        represented by ``caller_module`` (to prevent infinite recursion in case of cyclic
        imports).
        """
        key = item
        if self.names:
            if not key in self.names:
                return False
            key = self.names[key]
        for d in self.dicts:
            if d is caller_module:
                continue
            if key in d:
                return True
        return False


class IOOperationContentDict:
    """
    Represents the contents of an IOOperation, which does not use dicts
    internally and therefore does not work with CombinedDicts.
    """

    def __init__(self, io_op: 'PythonIOOperation') -> None:
        self.io_op = io_op

    def __contains__(self, item):
        result = self.io_op.get_variable(item)
        return result is not None

    def __getitem__(self, item):
        result = self.io_op.get_variable(item)
        return result


class ModuleDictView:
    """
    A view of the given aspect (e.g. 'functions') of the given module with the
    given renamings.
    """
    def __init__(self, names: List[Tuple[str, str]], module: 'PythonModule',
                 field: str):
        self.names = names
        self.module = module
        self.field = field
        self._dict = None

    def initialize(self) -> None:
        """
        We do this lazily because included module information may not be
        complete when we create this object.
        """
        modules = self.module.get_included_modules((), include_global=False)
        dicts = [getattr(m, self.field) for m in modules]
        self._dict = CombinedDict(self.names, dicts)

    def __contains__(self, item):
        if self._dict is None:
            self.initialize()
        return self._dict.contains(item, self)

    def __getitem__(self, item):
        if self._dict is None:
            self.initialize()
        return self._dict[item]


class PythonModuleView:
    """
    A view of a PythonModule that contains only the imported parts of an
    actual PythonModule, possibly with some renamings.
    """
    def __init__(self, module: 'PythonModule', names: List[Tuple[str, str]]):
        self.module = copy.copy(module)
        self.original_module = module
        for field in ['functions', 'methods', 'static_methods', 'namespaces',
                      'predicates', 'classes', 'global_vars', 'io_operations']:
            lazy_dict = ModuleDictView(names, module, field)
            setattr(self.module, field, lazy_dict)

    def get_contents(self, only_top: bool):
        return self.module.get_contents(only_top)

    def get_included_modules(self, exclude: Set['PythonModule'],
                             include_global: bool=True) -> List['PythonModule']:
        if self.module in exclude:
            return []
        return self.module.get_included_modules(exclude, include_global)

    def get_type(self, prefixes, name, previous=()):
        return self.module.get_type(prefixes, name, previous)

    def get_func_type(self, prefix):
        return self.module.get_func_type(prefix)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""
Monkey-patches member access resolution for union types so that no error
message is generated when accessing members of optional types, and type
inference and checking can proceed normally.
"""
import mypy.checkmember

from functools import wraps


def remove_none(f):
    @wraps(f)
    def wrapper(*args, **varargs):
        if isinstance(args[1], mypy.types.UnionType):
            args = list(args)
            members = [t for t in args[1].items
                       if not isinstance(t, mypy.types.NoneTyp)]
            new_type = mypy.types.UnionType.make_simplified_union(members)
            args[1] = new_type
            args = tuple(args)
        res = f(*args, **varargs)
        return res
    return wrapper

patched_access = remove_none(mypy.checkmember.analyze_member_access)
mypy.checkmember.analyze_member_access = patched_access



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Reference typed Silver expressions."""


import ast

from nagini_translation.lib.constants import PRIMITIVES
from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.silver_nodes.expression import Expression
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)


class RefExpression(Expression):   # pylint: disable=abstract-method
    """A base class for all reference expressions."""

    def __eq__(self, other) -> 'Expression':
        if other is None:
            other = Null()
        return super().__eq__(other)

    def __ne__(self, other) -> 'Expression':
        if other is None:
            other = Null()
        return super().__ne__(other)


class Null(RefExpression):
    """A null reference."""

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return translator.viper.NullLit(position, info)


class RefExpr(RefExpression):
    """A wrapper around an already-translated Silver expression of type Ref."""

    def __init__(self, e: Expr) -> None:
        self.e = e

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return self.e


class RefVar(RefExpression):
    """A reference to a Ref-typed variable."""

    def __init__(self, var: PythonVar) -> None:
        self._var = var

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        assert self._var.decl.typ() == translator.viper.Ref
        return self._var.ref()


class PythonRefExpression(RefExpression):
    """An reference expression represented by Python reference expression."""

    def __init__(self, node: ast.expr) -> None:
        self._node = node

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        stmt, expr = translator.translate_expr(
            self._node, ctx, impure=False)
        assert not stmt
        typ = translator.get_type(self._node, ctx)
        assert typ.name not in PRIMITIVES
        return expr



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Wait-level translator."""


import ast
import operator

from typing import Union

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.program_nodes import PythonVar
from nagini_translation.lib.typedefs import (
    StmtsAndExpr,
)
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.abstract import TranslatorConfig
from nagini_translation.translators.common import CommonTranslator
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)


class WaitLevelTranslator(CommonTranslator):
    """Class for translating wait-levels."""

    def __init__(self, config: TranslatorConfig, jvm: JVM, source_file: str,
                 type_info: TypeInfo, viper_ast: ViperAST,
                 obligation_manager: ObligationManager) -> None:
        super().__init__(config, jvm, source_file, type_info, viper_ast)
        self._obligation_manager = obligation_manager

    def is_wait_level_comparison(self, node: ast.Compare,
                                 ctx: Context) -> bool:
        """Check if we are comparing with ``WaitLevel``.

        Allowed forms are:

        1.  ``WaitLevel() < Level(l)``
        2.  ``Level(l1) < Level(l2)``
        """
        return (len(node.ops) == 1 and
                isinstance(node.ops[0], ast.Lt) and
                len(node.comparators) == 1 and
                isinstance(node.left, ast.Call) and
                isinstance(node.left.func, ast.Name) and
                (node.left.func.id in ('WaitLevel', 'Level')) and
                isinstance(node.comparators[0], ast.Call) and
                isinstance(node.comparators[0].func, ast.Name) and
                node.comparators[0].func.id == 'Level')

    def translate_wait_level_comparison(self, node: ast.Compare,
                                        ctx: Context) -> StmtsAndExpr:
        """Translate wait-level comparison.

        Comparisons can be in one of the following forms:

        1.  ``WaitLevel() < Level()``
        2.  ``Level() < Level()``
        """
        assert self.is_wait_level_comparison(node, ctx)
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)

        expr_right = self._translate_level(node.comparators[0])

        if node.left.func.id == 'Level':
            expr_left = self._translate_level(node.left)
            expression = sil.PermLtCmp(expr_left, expr_right)
        elif ctx.obligation_context.is_translating_loop():
            expression = self._translate_waitlevel_loop(
                expr_right, node, ctx)
        else:
            expression = self._translate_waitlevel_method(
                expr_right, node, ctx)
        return ([], expression.translate(self, ctx, position, info))

    def _translate_level(self, node: ast.Call) -> sil.PermExpression:
        """Translate a call to ``Level``."""
        assert len(node.args) == 1
        arg = sil.PythonRefExpression(node.args[0])
        return self.create_level_call(arg)

    def create_level_call(self, expr: sil.RefExpression) -> sil.PermExpression:
        return sil.PermCall('Level', [sil.CallArg('r', sil.REF, expr)])

    def _create_level_op(
            self, expr: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context,
            oper: Union[operator.le, operator.lt]) -> sil.BoolExpression:
        obligation = self._obligation_manager.must_release_obligation
        fields = obligation.create_fields_untranslated()
        var = ctx.current_function.create_variable(
            '_r', ctx.module.global_module.classes['object'],
            self.translator, local=False)
        op = oper(self.create_level_call(sil.RefVar(var)), expr)
        for_perms = [sil.ForPerm(var.sil_name, f, op) for f in fields]
        return sil.BigAnd(for_perms + [oper(residue_level_var, expr)])

    def create_level_below(
            self, expr: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        return self._create_level_op(expr, residue_level_var, ctx, operator.lt)

    def _create_level_below_equal(
            self, expr: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        return self._create_level_op(expr, residue_level_var, ctx, operator.le)

    def initialize_current_wait_level(
            self, current_wait_level: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        return sil.InhaleExhale(
            self._create_level_below_equal(
                current_wait_level,
                residue_level_var,
                ctx),
            sil.TrueLit())

    def _create_level_below_inex(
            self, guard: sil.BoolExpression, expr: sil.PermExpression,
            level: PythonVar, ctx: Context) -> sil.BoolExpression:
        return sil.InhaleExhale(
            sil.TrueLit(),
            sil.Implies(
                guard,
                self.create_level_below(
                    expr,
                    sil.PermVar(level),
                    ctx)))

    def _translate_waitlevel_loop(
            self, expr: sil.PermExpression, node: ast.Compare,
            ctx: Context) -> sil.BoolExpression:
        """Translate ``WaitLevel() < Level()`` in loop invariant."""
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)

        obligation_info = ctx.obligation_context.current_loop_info
        guard = obligation_info.get_wait_level_guard(node.left)
        context_info = ctx.obligation_context.get_surrounding_loop_info()
        if context_info:
            context_residue_level = context_info.residue_level
        else:
            method_info = ctx.current_function.obligation_info
            context_residue_level = method_info.residue_level
        body_inhale = sil.Inhale(
            sil.Implies(
                guard,
                sil.PermVar(obligation_info.residue_level) < expr))
        obligation_info.prepend_body(
            body_inhale.translate(self, ctx, position, info))
        body_exhale = sil.Implies(
            sil.Not(sil.BoolVar(obligation_info.loop_check_before_var)),
            self._create_level_below_inex(
                guard, expr, obligation_info.residue_level, ctx))
        obligation_info.add_invariant(
            body_exhale.translate(self, ctx, position, info))
        context_inhale = sil.Inhale(
            sil.Implies(
                guard,
                sil.PermVar(context_residue_level) < expr))
        obligation_info.append_after_loop(
            context_inhale.translate(self, ctx, position, info))
        context_exhale = sil.Implies(
            sil.BoolVar(obligation_info.loop_check_before_var),
            self._create_level_below_inex(
                guard, expr, context_residue_level, ctx))
        obligation_info.add_invariant(
            context_exhale.translate(self, ctx, position, info))
        return sil.TrueLit()

    def _translate_waitlevel_method(
            self, expr: sil.PermExpression, node: ast.Compare,
            ctx: Context) -> sil.BoolExpression:
        """Translate ``WaitLevel() < Level()`` in method contract."""
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)

        obligation_info = ctx.actual_function.obligation_info
        guard = obligation_info.get_wait_level_guard(node.left)
        exhale = self._create_level_below_inex(
            guard, expr, obligation_info.residue_level, ctx)
        translated_exhale = exhale.translate(self, ctx, position, info)
        if ctx.ignore_waitlevel_constraints:
            return sil.TrueLit()
        if ctx.obligation_context.is_translating_posts:
            obligation_info.add_postcondition(translated_exhale)
            caller_inhale = sil.InhaleExhale(
                sil.PermVar(obligation_info.current_wait_level) < expr,
                sil.TrueLit())
            return caller_inhale
        else:
            obligation_info.add_precondition(translated_exhale)
            body_inhale = sil.InhaleExhale(
                sil.PermVar(obligation_info.residue_level) < expr,
                sil.TrueLit())
            return body_inhale



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Base classes for all other nodes."""

# pragma pylint: disable=abstract-method


import abc

from nagini_translation.lib.typedefs import Info
from nagini_translation.lib.typedefs import Node as NodeType
from nagini_translation.lib.typedefs import Position


class Node(abc.ABC):
    """A base class for all nodes."""

    @abc.abstractmethod
    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> NodeType:
        """Translate to Silver."""


class Statement(Node):
    """A base class for all statements."""


class Expression(Node):
    """A base class for all expressions."""

    @abc.abstractmethod
    def __eq__(self, other: 'Expression') -> 'Expression':
        """Check equality."""

    @abc.abstractmethod
    def __ne__(self, other: 'Expression') -> 'Expression':
        """Check inequality."""


__all__ = (
    'Node',
    'Statement',
    'Expression',
)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import argparse
import astunparse
import inspect
import json
import logging
import os
import re
import time
import traceback

# These imports monkey-patch mypy and should happen as early as possible.
import nagini_translation.mypy_patches.column_info_patch
import nagini_translation.mypy_patches.optional_patch


from jpype import JavaException
from nagini_translation.analyzer import Analyzer
from nagini_translation.sif_translator import SIFTranslator
from nagini_translation.lib import config
from nagini_translation.lib.constants import DEFAULT_SERVER_SOCKET
from nagini_translation.lib.errors import error_manager
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.typedefs import Program
from nagini_translation.lib.typeinfo import TypeException, TypeInfo
from nagini_translation.lib.util import (
    ConsistencyException,
    InvalidProgramException,
    UnsupportedException,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.sif.lib.util import (
    configure_mpp_transformation,
    set_all_low_methods,
    set_preserves_low_methods
)
from nagini_translation.sif.lib.viper_ast_extended import ViperASTExtended
from nagini_translation.translator import Translator
from nagini_translation.verifier import (
    Carbon,
    get_arp_plugin,
    Silicon,
    VerificationResult,
    ViperVerifier
)
from typing import Set


TYPE_ERROR_PATTERN = r"^(?P<file>.*):(?P<line>\d+): error: (?P<msg>.*)$"
TYPE_ERROR_MATCHER = re.compile(TYPE_ERROR_PATTERN)


def parse_sil_file(sil_path: str, jvm):
    parser = getattr(getattr(jvm.viper.silver.parser, "FastParser$"), "MODULE$")
    assert parser
    with open(sil_path, 'r') as file:
        text = file.read()
    path = jvm.java.nio.file.Paths.get(sil_path, [])
    none = getattr(getattr(jvm.scala, 'None$'), 'MODULE$')
    parsed = parser.parse(text, path, none)
    assert (isinstance(parsed, getattr(jvm.fastparse.core,
                                       'Parsed$Success')))
    parse_result = parsed.value()
    parse_result.initProperties()
    resolver = jvm.viper.silver.parser.Resolver(parse_result)
    resolved = resolver.run()
    resolved = resolved.get()
    translator = jvm.viper.silver.parser.Translator(resolved, False)
    # Reset messages in global Consistency object. Otherwise, left-over
    # translation errors from previous translations prevent loading of the
    # built-in silver files.
    jvm.viper.silver.ast.utility.Consistency.resetMessages()
    program = translator.translate()
    return program.get()


def load_sil_files(jvm: JVM, sif: bool = False):
    current_path = os.path.dirname(inspect.stack()[0][1])
    if sif:
        resources_path = os.path.join(current_path, 'sif', 'resources')
    else:
        resources_path = os.path.join(current_path, 'resources')
    return parse_sil_file(os.path.join(resources_path, 'all.sil'), jvm)


def translate(path: str, jvm: JVM, selected: Set[str] = set(),
              sif: bool = False, arp: bool = False, ignore_global: bool = False,
              reload_resources: bool = False, verbose: bool = False) -> Program:
    """
    Translates the Python module at the given path to a Viper program
    """
    path = os.path.abspath(path)
    error_manager.clear()
    current_path = os.path.dirname(inspect.stack()[0][1])
    resources_path = os.path.join(current_path, 'resources')

    if sif:
        viper_ast = ViperASTExtended(jvm, jvm.java, jvm.scala, jvm.viper, path)
    else:
        viper_ast = ViperAST(jvm, jvm.java, jvm.scala, jvm.viper, path)
    if not viper_ast.is_available():
        raise Exception('Viper not found on classpath.')
    if sif and not viper_ast.is_extension_available():
        raise Exception('Viper AST SIF extension not found on classpath.')
    types = TypeInfo()
    type_correct = types.check(path)
    if not type_correct:
        return None

    analyzer = Analyzer(types, path, selected)
    main_module = analyzer.module
    with open(os.path.join(resources_path, 'preamble.index'), 'r') as file:
        analyzer.add_native_silver_builtins(json.loads(file.read()))

    main_module.add_builtin_vars()
    collect_modules(analyzer, path)
    if sif:
        translator = SIFTranslator(jvm, path, types, viper_ast)
    else:
        translator = Translator(jvm, path, types, viper_ast)
    analyzer.process(translator)
    if 'sil_programs' not in globals() or reload_resources:
        global sil_programs
        sil_programs = load_sil_files(jvm, sif)
    modules = [main_module.global_module] + list(analyzer.modules.values())
    prog = translator.translate_program(modules, sil_programs, selected,
                                        arp=arp, ignore_global=ignore_global)
    if sif:
        set_all_low_methods(jvm, viper_ast.all_low_methods)
        set_preserves_low_methods(jvm, viper_ast.preserves_low_methods)
    if verbose:
        print('Translation successful.')
    if sif:
        configure_mpp_transformation(jvm,
                                     ctrl_opt=True,
                                     seq_opt=True,
                                     act_opt=True,
                                     func_opt=True)
        prog = jvm.viper.silver.sif.SIFExtendedTransformer.transform(prog, False)
        if verbose:
            print('Transformation to MPP successful.')
    if arp:
        prog = get_arp_plugin(jvm).before_verify(prog)
        if verbose:
            print('ARP transformation successful.')
    # Run consistency check in translated AST
    consistency_errors = viper_ast.to_list(prog.checkTransitively())
    for error in consistency_errors:
        print(error.toString())
    if consistency_errors:
        print(prog)
        raise ConsistencyException('consistency.error')
    return prog


def collect_modules(analyzer: Analyzer, path: str) -> None:
    """
    Starting from the main module, finds all imports and sets up all modules
    for them.
    """
    analyzer.module_index = 0
    analyzer.collect_imports(path)

    analyzer.analyze()

    # Carry out all tasks that were deferred to the end of the analysis.
    for task in analyzer.deferred_tasks:
        task()


def verify(prog: 'viper.silver.ast.Program', path: str,
           jvm: JVM, backend=ViperVerifier.silicon, arp=False) -> VerificationResult:
    """
    Verifies the given Viper program
    """
    try:
        if backend == ViperVerifier.silicon:
            verifier = Silicon(jvm, path)
        elif backend == ViperVerifier.carbon:
            verifier = Carbon(jvm, path)
        vresult = verifier.verify(prog, arp=arp)
        return vresult
    except JavaException as je:
        print(je.stacktrace())
        traceback.print_exc()


def _parse_log_level(log_level_string: str) -> int:
    """ Parses the log level provided by the user.
    """
    LOG_LEVELS = ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']

    log_level_string_upper = log_level_string.upper()
    if log_level_string_upper in LOG_LEVELS:
        return getattr(logging, log_level_string_upper, logging.WARNING)
    else:
        msg = 'Invalid logging level {0} (expected one of: {1})'.format(
            log_level_string,
            LOG_LEVELS)
        raise argparse.ArgumentTypeError(msg)


def main() -> None:
    """ Entry point for the translator.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'python_file',
        help='Python file to verify')
    parser.add_argument(
        '--viper-jar-path',
        help='Java CLASSPATH that includes Viper class files',
        default=config.classpath)
    parser.add_argument(
        '--boogie',
        help='path to Boogie executable',
        default=config.boogie_path)
    parser.add_argument(
        '--z3',
        help='path to Z3 executable',
        default=config.z3_path)
    parser.add_argument(
        '--mypy-path',
        help='mypy path',
        default=config.mypy_path)
    parser.add_argument(
        '--print-silver',
        action='store_true',
        help='print generated Silver program')
    parser.add_argument(
        '--write-silver-to-file',
        default=None,
        help='write generated Silver program to specified file')
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="increase output verbosity")
    parser.add_argument(
        '--verifier',
        help='verifier to be used (carbon or silicon)',
        default='silicon')
    parser.add_argument(
        '--sif',
        action='store_true',
        help='verify secure information flow')
    parser.add_argument(
        '--show-viper-errors',
        action='store_true',
        help='show Viper-level error messages if no Python errors are available')
    parser.add_argument(
        '--arp',
        action='store_true',
        help='Use Abstract Read Permissions')
    parser.add_argument(
        '--log',
        type=_parse_log_level,
        help='log level',
        default='WARNING')
    parser.add_argument(
        '--benchmark',
        type=int,
        help=('run verification the given number of times to benchmark '
              'performance'),
        default=-1)
    parser.add_argument(
        '--ide-mode',
        action='store_true',
        help='Output errors in IDE format')
    parser.add_argument(
        '--select',
        default=None,
        help='select specific methods or classes to verify, separated by commas'
    )
    parser.add_argument(
        '--ignore-global',
        action='store_true',
        help='do not verify the the top level program (global statements)'
    )
    parser.add_argument(
        '--server',
        action='store_true',
        help='Start Nagini server'
    )
    args = parser.parse_args()

    config.classpath = args.viper_jar_path
    config.boogie_path = args.boogie
    config.z3_path = args.z3
    config.mypy_path = args.mypy_path
    config.set_verifier(args.verifier)

    if not config.classpath:
        parser.error('missing argument: --viper-jar-path')
    if not config.z3_path:
        parser.error('missing argument: --z3')
    if args.verifier == 'carbon' and not config.classpath:
        parser.error('missing argument: --boogie')

    logging.basicConfig(level=args.log)

    os.environ['MYPYPATH'] = config.mypy_path
    jvm = JVM(config.classpath)
    if args.server:
        import zmq
        context = zmq.Context()
        socket = context.socket(zmq.REP)
        socket.bind(DEFAULT_SERVER_SOCKET)
        global sil_programs
        sil_programs = load_sil_files(jvm, args.sif)

        while True:
            file = socket.recv_string()
            response = ['']

            def add_response(part):
                response[0] = response[0] + '\n' + part

            translate_and_verify(file, jvm, args, add_response, arp=args.arp)
            socket.send_string(response[0])
    else:
        translate_and_verify(args.python_file, jvm, args, arp=args.arp)


def translate_and_verify(python_file, jvm, args, print=print, arp=False):
    try:
        start = time.time()
        selected = set(args.select.split(',')) if args.select else set()
        prog = translate(python_file, jvm, selected, args.sif,
                         ignore_global=args.ignore_global, arp=arp, verbose=args.verbose)
        if args.print_silver:
            if args.verbose:
                print('Result:')
            print(str(prog))
        if args.write_silver_to_file:
            with open(args.write_silver_to_file, 'w') as fp:
                fp.write(str(prog))
        if args.verifier == 'silicon':
            backend = ViperVerifier.silicon
        elif args.verifier == 'carbon':
            backend = ViperVerifier.carbon
        else:
            raise ValueError('Unknown verifier specified: ' + args.verifier)
        if args.benchmark >= 1:
            print("Run, Total, Start, End, Time".format())
            for i in range(args.benchmark):
                start = time.time()
                prog = translate(python_file, jvm, selected, args.sif, arp=arp)
                vresult = verify(prog, python_file, jvm, backend=backend, arp=arp)
                end = time.time()
                print("{}, {}, {}, {}, {}".format(
                    i, args.benchmark, start, end, end - start))
        else:
            vresult = verify(prog, python_file, jvm, backend=backend, arp=arp)
        if args.verbose:
            print("Verification completed.")
        print(vresult.to_string(args.ide_mode, args.show_viper_errors))
        duration = '{:.2f}'.format(time.time() - start)
        print('Verification took ' + duration + ' seconds.')
    except (TypeException, InvalidProgramException, UnsupportedException) as e:
        print("Translation failed")
        if isinstance(e, (InvalidProgramException, UnsupportedException)):
            if isinstance(e, InvalidProgramException):
                issue = 'Invalid program: '
                if e.message:
                    issue += e.message
                else:
                    issue += e.code
            else:
                issue = 'Not supported: '
                if e.args[0]:
                    issue += e.args[0]
                else:
                    issue += astunparse.unparse(e.node)
            line = str(e.node.lineno)
            col = str(e.node.col_offset)
            print(issue + ' (' + python_file + '@' + line + '.' + col + ')')
        if isinstance(e, TypeException):
            for msg in e.messages:
                parts = TYPE_ERROR_MATCHER.match(msg)
                if parts:
                    parts = parts.groupdict()
                    file = parts['file']
                    if file == '__main__':
                        file = python_file
                    msg = parts['msg']
                    line = parts['line']
                    print('Type error: ' + msg + ' (' + file + '@' + line + '.0)')
                else:
                    print(msg)
    except ConsistencyException as e:
        print(e.message + ': Translated AST contains inconsistencies.')

    except JavaException as e:
        print(e.stacktrace())
        raise e


if __name__ == '__main__':
    main()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from abc import ABCMeta
from nagini_translation.lib.constants import (
    ARBITRARY_BOOL_FUNC,
    ASSERTING_FUNC,
    COMBINE_NAME_FUNC,
    DICT_TYPE,
    INT_TYPE,
    IS_DEFINED_FUNC,
    LIST_TYPE,
    MAIN_METHOD_NAME,
    MAY_SET_PRED,
    NAME_DOMAIN,
    PRIMITIVE_BOOL_TYPE,
    PRIMITIVE_INT_TYPE,
    RANGE_TYPE,
    PSEQ_TYPE,
    SET_TYPE,
    SINGLE_NAME,
    UNION_TYPE,
)
from nagini_translation.lib.context import Context
from nagini_translation.lib.errors import rules
from nagini_translation.lib.program_nodes import (
    chain_cond_exp,
    chain_if_stmts,
    OptionalType,
    PythonClass,
    PythonField,
    PythonIOOperation,
    PythonMethod,
    PythonModule,
    PythonNode,
    PythonType,
    PythonVar,
    toposort_classes,
    UnionType,
)
from nagini_translation.lib.resolver import get_target as do_get_target
from nagini_translation.lib.typedefs import (
    Expr,
    FuncApp,
    Info,
    Position,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.util import (
    get_surrounding_try_blocks,
    InvalidProgramException,
    UnsupportedException
)
from nagini_translation.translators.abstract import AbstractTranslator
from typing import List, Tuple, Union


class CommonTranslator(AbstractTranslator, metaclass=ABCMeta):
    """
    Abstract class which all specialized translators extend. Provides some
    functionality which is needed by many or all specialized translators.
    """

    def translate_generic(self, node: ast.AST, ctx: Context) -> None:
        """
        Visitor that is used if no other visitor is implemented.
        Simply raises an exception.
        """
        raise UnsupportedException(node)

    def translate_block(self, stmtlist: List['silver.ast.Stmt'],
                        position: 'silver.ast.Position',
                        info: 'silver.ast.Info') -> Stmt:
        """
        Wraps a (Python) list of (Viper) statements into a Viper block
        """
        body = []
        for stmt in stmtlist:
            body.append(stmt)
        return self.viper.Seqn(body, position, info)

    def convert_to_type(self, e: Expr, target_type, ctx: Context,
                        node: ast.AST = None) -> Expr:
        """
        Converts expression ``e`` to the Viper type ``target_type`` if said
        type is Ref, Bool or Int.
        """
        result = e
        if target_type == self.viper.Ref:
            result = self.to_ref(e, ctx)
        elif target_type == self.viper.Bool:
            result = self.to_bool(e, ctx, node)
        elif target_type == self.viper.Int:
            result = self.to_int(e, ctx)
        return result

    def _is_pure(self, e: Expr) -> bool:
        e = self.unwrap(e)
        if isinstance(e, (self.viper.ast.And, self.viper.ast.Or)):
            return self._is_pure(e.left()) and self._is_pure(e.right())
        return e.isPure()

    def to_type(self, e: Expr, t, ctx) -> Expr:
        if t is self.viper.Ref:
            return self.to_ref(e, ctx)
        if t is self.viper.Int:
            return self.to_int(e, ctx)
        if t is self.viper.Bool:
            return self.to_bool(e, ctx)
        return e

    def to_ref(self, e: Expr, ctx: Context) -> Expr:
        """
        Converts the given expression to an expression of the Silver type Ref
        if it isn't already, either by boxing a primitive or undoing a
        previous unboxing operation.
        """
        # Avoid wrapping non-pure expressions (leads to errors within Silver's
        # Consistency object)
        if not self._is_pure(e):
            return e
        result = e
        if e.typ() == self.viper.Int:
            if (isinstance(e, self.viper.ast.FuncApp) and
                    e.funcname() == 'int___unbox__'):
                result = e.args().head()
            else:
                prim_int = ctx.module.global_module.classes[PRIMITIVE_INT_TYPE]
                result = self.get_function_call(prim_int, '__box__',
                                                [result], [None], None, ctx,
                                                position=e.pos())
        elif e.typ() == self.viper.Bool:
            if (isinstance(e, self.viper.ast.FuncApp) and
                    e.funcname() == 'bool___unbox__'):
                result = e.args().head()
            else:
                prim_bool = ctx.module.global_module.classes[PRIMITIVE_BOOL_TYPE]
                result = self.get_function_call(prim_bool, '__box__',
                                                [result], [None], None, ctx,
                                                position=e.pos())
        return result

    def to_bool(self, e: Expr, ctx: Context, node: ast.AST = None) -> Expr:
        """
        Converts the given expression to an expression of the Silver type Bool
        if it isn't already, either by calling __bool__ on an object and
        possibly unboxing the result, or by undoing a previous boxing operation.
        """
        # Avoid wrapping non-pure expressions (leads to errors within Silver's
        # Consistency object)
        if not self._is_pure(e):
            return e
        if e.typ() == self.viper.Bool:
            return e
        if e.typ() != self.viper.Ref:
            e = self.to_ref(e, ctx)
        if (isinstance(e, self.viper.ast.FuncApp) and
                e.funcname() == '__prim__bool___box__'):
            return e.args().head()
        result = e
        call_bool = True
        if node:
            node_type = self.get_type(node, ctx)
            if node_type.name == 'bool':
                call_bool = False
            if call_bool:
                result = self.get_function_call(node_type, '__bool__',
                                                [result], [None], node, ctx,
                                                position=e.pos())
        if result.typ() != self.viper.Bool:
            bool_type = ctx.module.global_module.classes['bool']
            result = self.get_function_call(bool_type, '__unbox__',
                                            [result], [None], node, ctx,
                                            position=e.pos())
        return result

    def to_int(self, e: Expr, ctx: Context) -> Expr:
        """
        Converts the given expression to an expression of the Silver type Int
        if it isn't already, either by unboxing a reference or undoing a
        previous boxing operation.
        """
        # Avoid wrapping non-pure expressions (leads to errors within Silver's
        # Consistency object)
        if not self._is_pure(e):
            return e
        if e.typ() == self.viper.Int:
            return e
        if e.typ() != self.viper.Ref:
            e = self.to_ref(e, ctx)
        if (isinstance(e, self.viper.ast.FuncApp) and
                    e.funcname() == '__prim__int___box__'):
            return e.args().head()
        result = e
        int_type = ctx.module.global_module.classes[INT_TYPE]
        result = self.get_function_call(int_type, '__unbox__',
                                        [result], [None], None, ctx,
                                        position=e.pos())
        return result

    def unwrap(self, e: Expr) -> Expr:
        if isinstance(e, self.viper.ast.FuncApp):
            if (e.funcname().endswith('__box__') or
                    e.funcname().endswith('__unbox__')):
                return e.args().head()
        return e

    def to_position(
            self, node: ast.AST, ctx: Context, error_string: str=None,
            rules: rules.Rules=None) -> 'silver.ast.Position':
        """
        Extracts the position from a node, assigns an ID to the node and stores
        the node and the position in the context for it.
        """
        return self.viper.to_position(node, ctx.position, error_string, rules,
                                      ctx.module.file)

    def no_position(self, ctx: Context, error_string: str=None,
            rules: rules.Rules=None) -> 'silver.ast.Position':
        return self.to_position(None, ctx, error_string, rules)

    def to_info(self, comments: List[str], ctx: Context) -> 'silver.ast.Info':
        """
        Wraps the given comments into an Info object.
        If ctx.info is set to override the given info, returns that.
        """
        if ctx.info is not None:
            return ctx.info
        if comments:
            return self.viper.SimpleInfo(comments)
        else:
            return self.viper.NoInfo

    def no_info(self, ctx: Context) -> 'silver.ast.Info':
        return self.to_info([], ctx)

    def normalize_type(self, typ: PythonType, ctx: Context) -> PythonType:
        """
        Normalizes a type, i.e., returns the actual NoneType if it's None,
        otherwise just returns the type.
        """
        if typ is None:
            return ctx.module.global_module.classes['NoneType']
        return typ

    def is_local_variable(self, var: PythonVar, ctx: Context) -> bool:
        """
        Assuming we are currently inside an impure method, checks if the given variable is
        a local variable in the Python program (i.e. not a parameter, not a variable that
        is quantified over).
        """
        if not ctx.actual_function:
            return False
        if var.name in ctx.actual_function.args:
            return False
        return var in ctx.actual_function.locals.values()

    def get_may_set_predicate(self, rec: Expr, field: PythonField, ctx: Context,
                              pos: Position = None) -> Expr:
        """
        Creates predicate instances representing the permissions to create the given
        field on the given receiver object.
        """
        if not pos:
            pos = self.no_position(ctx)
        info = self.no_info(ctx)
        full_perm = self.viper.FullPerm(pos, info)
        id = self.viper.IntLit(self._get_string_value(field.sil_name), pos, info)
        pred = self.viper.PredicateAccess([rec, id], MAY_SET_PRED, pos, info)
        pred_acc = self.viper.PredicateAccessPredicate(pred, full_perm, pos, info)
        return pred_acc

    def check_var_defined(self, target: PythonVar, position: Position,
                        info: Info) -> Expr:
        id = self.viper.IntLit(self._get_string_value(target.sil_name), position, info)
        id_param_decl = self.viper.LocalVarDecl('id', self.viper.Int, position, info)
        is_defined = self.viper.FuncApp(IS_DEFINED_FUNC, [id], position, info,
                                        self.viper.Bool, [id_param_decl])
        return is_defined

    def set_var_defined(self, target: PythonVar, position: Position,
                        info: Info) -> Stmt:
        """
        Returns an inhale which assumes that the given local variable is now defined.
        """
        is_defined = self.check_var_defined(target, position, info)
        return self.viper.Inhale(is_defined, position, info)

    def set_global_defined(self, declaration: PythonNode, module: PythonModule,
                           node: ast.AST, ctx: Context) -> Stmt:
        """
        Returns a statement that sets the name of the given declaration to be defined
        in the given module.
        """
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        module_set = module.names_var[1]
        decl_id = self.viper.IntLit(self._get_string_value(declaration.name), pos,
                                    info)
        return self._set_global_defined(decl_id, module_set, pos, info)

    def _set_global_defined(self, decl_int: Expr, module_var: Expr, pos: Position,
                            info: Info) -> Stmt:
        """
        Returns a statement that sets the name of represented by the integer decl_int to
        be defined in the given set of names.
        """
        if decl_int.typ() == self.viper.Int:
            decl_int = self.viper.DomainFuncApp(SINGLE_NAME, [decl_int], self.name_type(),
                                                pos, info, NAME_DOMAIN)
        new_set = self.viper.ExplicitSet([decl_int], pos, info)
        union = self.viper.AnySetUnion(module_var, new_set, pos, info)
        return self.viper.LocalVarAssign(module_var, union, pos, info)

    def name_type(self) -> 'silver.ast.DomainType':
        """
        The Silver type of global names, for which one can check if they are defined or
        not.
        """
        return self.viper.DomainType(NAME_DOMAIN, {}, [])

    def _is_defined(self, name: Expr, module: Expr, pos: Position, info: Info) -> Expr:
        """
        Returns an expression that is true iff the name represented by the given
        expression is defined in the module represented by the other expression.
        """
        name_type = self.viper.DomainType(NAME_DOMAIN, {}, [])
        if name.typ() == self.viper.Int:
            boxed_name = self.viper.DomainFuncApp(SINGLE_NAME, [name], name_type, pos,
                                                  info, NAME_DOMAIN)
        else:
            boxed_name = name
        return self.viper.AnySetContains(boxed_name, module, pos, info)

    def _combine_names(self, prefix: Expr, name: Expr, pos: Position, info: Info) -> Expr:
        """
        Returns an expression that combines the prefix-name and the name to a new name
        that represents 'prefix.name'.
        """
        name_type = self.viper.DomainType(NAME_DOMAIN, {}, [])
        if name.typ() == self.viper.Int:
            boxed_name = self.viper.DomainFuncApp(SINGLE_NAME, [name], name_type, pos,
                                                  info, NAME_DOMAIN)
        else:
            boxed_name = name
        if prefix.typ() == self.viper.Int:
            boxed_prefix = self.viper.DomainFuncApp(SINGLE_NAME, [prefix], name_type, pos,
                                                    info, NAME_DOMAIN)
        else:
            boxed_prefix = prefix
        return self.viper.DomainFuncApp(COMBINE_NAME_FUNC, [boxed_prefix, boxed_name],
                                        name_type, pos, info, NAME_DOMAIN)

    def extract_identifiers(self, ref: ast.AST, pos: Position,
                            info: Info) -> List[Expr]:
        """
        Returns a list containing all names contained by the given reference.
        """
        res = []
        if isinstance(ref, ast.Subscript):
            if isinstance(ref.value, ast.Name) and ref.value.id in ('Optional', 'Union'):
                if not isinstance(ref.slice.value, ast.Tuple):
                    return self.extract_identifiers(ref.slice.value, pos, info)
                for e in ref.slice.value.elts:
                    res.extend(self.extract_identifiers(e, pos, info))
                return res

        decl_id = None
        for name in reversed(self._get_name_parts(ref)):
            current = self.viper.IntLit(self._get_string_value(name), pos,
                                        info)
            if decl_id is None:
                decl_id = current
            else:
                decl_id = self._combine_names(current, decl_id, pos, info)
        if decl_id:
            return [decl_id]
        return []

    def _get_global_definedness_conditions(self, declaration: PythonNode,
                                           module: PythonModule, ref_node: ast.AST,
                                           ctx: Context) -> Tuple[Expr, Expr]:
        """
        Returns two boolean expressions that represent 1) if the name of the given
        declaration is defined in the given module, and 2) if all dependencies of the
        given declaration are currently defined.
        """
        msg = 'Name "' + declaration.name + '" is defined'
        pos = self.to_position(ref_node, ctx, error_string=msg)
        info = self.no_info(ctx)
        module_set = module.names_var[1]
        decl_ids = self.extract_identifiers(ref_node, pos, info)
        contains = self.viper.TrueLit(pos, info)
        for decl_id in decl_ids:
            contains = self.viper.And(contains, self._is_defined(decl_id, module_set, pos,
                                                                 info), pos, info)
        deps = set()
        if isinstance(declaration, (PythonMethod, PythonClass)):
            called = declaration
            if isinstance(called, PythonClass):
                called = called.get_method('__init__')
            if called:
                called.add_all_call_deps(deps)
        msg = 'all dependencies of "' + declaration.name + '" are defined'
        pos = self.to_position(ref_node, ctx, error_string=msg)
        deps_defined = self.viper.TrueLit(pos, info)
        for ref, decl, mod, *conds in deps:
            module_set = mod.names_var[1]
            decl_ids = self.extract_identifiers(ref, pos, info)
            for decl_id in decl_ids:
                contains_dep = self._is_defined(decl_id, module_set, pos, info)
                for cond in conds:
                    # Iterate over conditions (PythonNodes); the dependency must be
                    # defined if all such PythonNodes are currently defined in their
                    # respective modules.
                    module_set = cond.module.names_var[1]
                    decl_id = self.viper.IntLit(self._get_string_value(cond.name), pos,
                                                info)
                    cond_contains = self._is_defined(decl_id, module_set, pos, info)
                    contains_dep = self.viper.Implies(cond_contains, contains_dep, pos,
                                                      info)
                deps_defined = self.viper.And(deps_defined, contains_dep, pos, info)

        return contains, deps_defined

    def _get_name_parts(self, node: ast.AST) -> List[str]:
        """
        Converts an AST node representing some kind of reference to a list of strings.
        """
        while isinstance(node, ast.Subscript):
            node = node.value
        if isinstance(node, ast.Name):
            return [node.id]
        if isinstance(node, ast.Attribute):
            pref = self._get_name_parts(node.value)
            return pref + [node.attr]
        if isinstance(node, ast.Str):
            return []
        return [node.name]

    def assert_global_defined(self, declaration: PythonNode, module: PythonModule,
                              ref_node: ast.AST, ctx: Context,
                              call_deps=True) -> List[Stmt]:
        """
        Creates assertions that check that the given declaration and all its dependencies
        are currently defined in the given module.
        """
        info = self.no_info(ctx)
        name, deps = self._get_global_definedness_conditions(declaration, module,
                                                             ref_node, ctx)
        msg = 'Name "' + declaration.name + '" is defined'
        pos = self.to_position(ref_node, ctx, error_string=msg)
        assert_name = self.viper.Assert(name, pos, info)
        if not call_deps:
            return [assert_name]
        msg = 'all dependencies of "' + declaration.name + '" are defined'
        pos = self.to_position(ref_node, ctx, error_string=msg)
        assert_deps = self.viper.Assert(deps, pos, info)
        return [assert_name, assert_deps]

    def wrap_global_defined_check(self, val: Expr, declaration: PythonNode,
                                  module: PythonModule, ref_node: ast.AST,
                                  ctx: Context) -> Expr:
        """
        Wraps the given expression into a new expression that checks that the given
        declaration and all its dependencies are currently defined in the given module.
        """
        info = self.no_info(ctx)
        msg = 'Name "' + declaration.name + '" is defined'
        pos = self.to_position(ref_node, ctx, error_string=msg,
                               rules=rules.GLOBAL_NAME_NOT_DEFINED)
        msg = 'all dependencies of "' + declaration.name + '" are defined'
        deps_pos = self.to_position(ref_node, ctx, error_string=msg,
                                    rules=rules.DEPENDENCIES_NOT_DEFINED)
        name, deps = self._get_global_definedness_conditions(declaration, module,
                                                             ref_node, ctx)
        assertion_param_decl = self.viper.LocalVarDecl('ass',
                                                       self.viper.Bool, pos,
                                                       info)
        var_param_decl = self.viper.LocalVarDecl('val', self.viper.Ref, pos, info)
        deps_func = self.viper.FuncApp(ASSERTING_FUNC, [val, deps], deps_pos, info,
                                       self.viper.Ref, [var_param_decl,
                                                        assertion_param_decl])
        name_func = self.viper.FuncApp(ASSERTING_FUNC, [deps_func, name], pos, info,
                                       self.viper.Ref, [var_param_decl,
                                                        assertion_param_decl])
        return name_func

    def is_main_method(self, ctx: Context) -> bool:
        """
        Checks if we are currently translating the 'main method', i.e., the global
        statements of the program.
        """
        if not ctx.current_function:
            return False
        return ctx.current_function.name == MAIN_METHOD_NAME

    def get_tuple_type_arg(self, arg: Expr, arg_type: PythonType, node: ast.AST,
                           ctx: Context) -> Expr:
        """
        Creates an expression of type PyType that represents the type of 'arg',
        to be handed to the constructor function for tuples. This is different
        than what's used elsewhere. For, e.g., Optional[NoneType, A, C], this
        will return
        arg == null ? NoneType : issubtype(typeof(arg), A) ? A : C
        """
        position = self.no_position(ctx)
        info = self.no_info(ctx)
        if arg_type.name == UNION_TYPE:
            first_arg = self.normalize_type(arg_type.type_args[0], ctx)
            result = self.type_factory.translate_type_literal(first_arg,
                                                              position, ctx)
            for option in arg_type.type_args[1:]:
                option = self.normalize_type(option, ctx)
                check = self.type_check(arg, option, position, ctx, False)
                type_lit = self.type_factory.translate_type_literal(option,
                                                                    position,
                                                                    ctx)
                result = self.viper.CondExp(check, type_lit, result, position,
                                            info)
            return result
        arg_type = self.normalize_type(arg_type, ctx)
        type_lit = self.type_factory.translate_type_literal(arg_type,
                                                            position, ctx)
        return type_lit

    def get_func_or_method_call(self, receiver: PythonType, func_name: str,
                                args: List[Expr], arg_types: List[Expr],
                                node: ast.AST, ctx: Context) -> StmtsAndExpr:
        if receiver.has_function(func_name):
            call = self.get_function_call(receiver, func_name, args, arg_types, node, ctx)
            return [], call
        method = receiver.get_method(func_name)
        if method:
            assert method.type
            target_var = ctx.actual_function.create_variable('target', method.type,
                                                             self.translator)
            val = target_var.ref(node, ctx)
            call = self.get_method_call(receiver, func_name, args, arg_types, [val], node,
                                        ctx)
            return call, val
        return [], None

    def get_quantifier_lhs(self, in_expr: Expr, dom_type: PythonType, dom_arg: Expr,
                           node: ast.AST, ctx: Context, position: Position,
                           force_trigger=False) -> Expr:
        """
        Returns a contains-expression representing whether in_expr is in dom_arg.
        To be used on the left hand side of quantifiers (and in the corresponding
        triggers):
        Forall(iter, lambda x: e)
        becomes
        forall x: <quantifier_lhs> ==> e
        Defaults to in_expr in type___sil_seq__, but used simpler expressions for known
        types to improve performance/triggering.
        """
        position = position if position else self.to_position(node, ctx)
        info = self.no_info(ctx)
        res = None
        if not (isinstance(dom_type, UnionType) or isinstance(dom_type, OptionalType)):
            if dom_type.name in (DICT_TYPE, SET_TYPE):
                contains_constructor = self.viper.AnySetContains
                if dom_type.name == DICT_TYPE:
                    set_ref = self.viper.SetType(self.viper.Ref)
                    field = self.viper.Field('dict_acc', set_ref, position, info)
                    res = self.viper.FieldAccess(dom_arg, field, position, info)
                if dom_type.name == SET_TYPE:
                    set_ref = self.viper.SetType(self.viper.Ref)
                    field = self.viper.Field('set_acc', set_ref, position, info)
                    res = self.viper.FieldAccess(dom_arg, field, position, info)
            if False and (dom_type.name == RANGE_TYPE and isinstance(node.func, ast.Name) and
                        node.func.id == 'range'):
                left = node.args[0]
                right = node.args[1]
                _, left_expr = self.translate_expr(left, ctx)
                _, right_expr = self.translate_expr(right, ctx)
                int_class = ctx.module.global_module.classes[INT_TYPE]
                left_bound = self.get_function_call(int_class, '__ge__',
                                                    [in_expr, left], [None, None],
                                                    node, ctx, position)
                right_bound = self.get_function_call(int_class, '__lt__',
                                                    [in_expr, right], [None, None],
                                                    node, ctx, position)
                if force_trigger:
                    return None
                else:
                    return self.viper.And(left_bound, right_bound, position, info)
        if res is None:
            contains_constructor = self.viper.SeqContains
            res = self.get_sequence(dom_type, dom_arg, None, node, ctx, position)
        return contains_constructor(in_expr, res, position, info)

    def get_sequence(self, receiver: PythonType, arg: Expr, arg_type: PythonType,
                     node: ast.AST, ctx: Context,
                     position: Position = None) -> Expr:
        """
        Returns a sequence (Viper type Seq[Ref]) representing the contents of arg.
        Defaults to type___sil_seq__, but used simpler expressions for known types
        to improve performance/triggering.
        """
        position = position if position else self.to_position(node, ctx)
        info = self.no_info(ctx)
        if not isinstance(receiver, UnionType) or isinstance(receiver, OptionalType):
            if receiver.name == LIST_TYPE:
                seq_ref = self.viper.SeqType(self.viper.Ref)
                field = self.viper.Field('list_acc', seq_ref, position, info)
                res = self.viper.FieldAccess(arg, field, position, info)
                return res
            if receiver.name == PSEQ_TYPE:
                if (isinstance(arg, self.viper.ast.FuncApp) and
                            arg.funcname() == 'PSeq___create__'):
                    args = self.viper.to_list(arg.args())
                    return args[0]
        return self.get_function_call(receiver, '__sil_seq__', [arg], [arg_type],
                                      node, ctx, position)

    def _get_function_call(self, receiver: PythonType,
                          func_name: str, args: List[Expr],
                          arg_types: List[PythonType], node: ast.AST,
                          ctx: Context,
                          position: Position = None) -> FuncApp:
        """
        Creates a function application of the function called func_name, with
        the given receiver and arguments. Boxes arguments if necessary, and
        unboxed the result if needed as well. This method only handles receivers
        of non-union types.
        """
        if receiver:
            target_cls = receiver
            func = target_cls.get_function(func_name)
        else:
            for container in ctx.module.get_included_modules():
                if func_name in container.functions:
                    func = container.functions[func_name]
                    break
        if not func:
            if receiver and target_cls.get_method(func_name):
                msg = 'Called method is expected to be pure: ' + func_name
                raise UnsupportedException(node, msg)
            raise InvalidProgramException(node, 'unknown.function.called')
        formal_args = []
        actual_args = []
        assert len(args) == len(func.get_args())
        for arg, param, type in zip(args, func.get_args(), arg_types):
            formal_args.append(param.decl)
            if param.type.name == '__prim__bool':
                actual_arg = self.to_bool(arg, ctx)
            elif param.type.name == '__prim__int':
                actual_arg = self.to_int(arg, ctx)
            else:
                actual_arg = self.to_ref(arg, ctx)
            actual_args.append(actual_arg)
        type = self.translate_type(func.type, ctx)
        sil_name = func.sil_name

        actual_position = position if position else self.to_position(node, ctx)
        call = self.viper.FuncApp(sil_name, actual_args,
                                  actual_position,
                                  self.no_info(ctx), type, formal_args)
        return call

    def get_function_call(self, receiver: PythonType,
                          func_name: str, args: List[Expr],
                          arg_types: List[PythonType], node: ast.AST,
                          ctx: Context,
                          position: Position = None) -> FuncApp:
        """
        Creates a function application of the function called func_name, with
        the given receiver and arguments. Boxes arguments if necessary, and
        unboxed the result if needed as well. When the receiver is of union
        type, a function call application is created for each type in the
        union with its respective guard.
        """
        if receiver and type(receiver) is UnionType:
            position = self.to_position(node, ctx) if position is None else position
            guarded_functions = []
            for cls in toposort_classes(receiver.get_types() - {None}):

                # Create guard checking if receiver is an instance of this class
                guard = self.type_check(args[0], cls, position, ctx)

                # Translate the function call on this particular receiver's class
                function = self._get_function_call(cls, func_name, args,
                                                   arg_types, node, ctx,
                                                   position)

                # Stores guard and translated function call as tuple in a list
                guarded_functions.append((guard, function))

            # Chain list of guard and function call tuples in an if-then-else
            # expression
            return chain_cond_exp(guarded_functions, self.viper, position,
                                  self.no_info(ctx), ctx)
        else:
            # Pass-through
            return self._get_function_call(receiver, func_name, args,
                                           arg_types, node, ctx, position)

    def _get_method_call(self, receiver: PythonType,
                        func_name: str, args: List[Expr],
                        arg_types: List[PythonType],
                        targets: List['silver.ast.LocalVarRef'],
                        node: ast.AST,
                        ctx: Context) -> List[Stmt]:
        """
        Creates a method call to the method called func_name, with the given
        receiver and arguments. Boxes arguments if necessary. This method only
        handles receivers of non-union types.
        """
        if receiver:
            target_cls = receiver
            func = target_cls.get_method(func_name)
        else:
            func = ctx.module.methods[func_name]
        if not func:
            raise InvalidProgramException(node, 'unknown.method.called')
        actual_args = []
        for arg, param, _ in zip(args, func.get_args(), arg_types):
            if param.type.name == PRIMITIVE_BOOL_TYPE:
                actual_arg = self.to_bool(arg, ctx)
            elif param.type.name == '__prim__int':
                actual_arg = self.to_int(arg, ctx)
            else:
                actual_arg = self.to_ref(arg, ctx)
            actual_args.append(actual_arg)
        sil_name = func.sil_name
        call = self.create_method_call_node(
            ctx, sil_name, actual_args, targets, self.to_position(node, ctx),
            self.no_info(ctx), target_method=func, target_node=node)
        return call

    def get_method_call(self, receiver: PythonType,
                        func_name: str, args: List[Expr],
                        arg_types: List[PythonType],
                        targets: List['silver.ast.LocalVarRef'],
                        node: ast.AST,
                        ctx: Context) -> List[Stmt]:
        """
        Creates a method call to the method called func_name, with the given
        receiver and arguments. Boxes arguments if necessary. When the receiver
        is of union type, a method call is created for each type in the union
        with its respective guard.
        """
        position = self.to_position(node, ctx)
        info = self.no_info(ctx)
        if receiver and type(receiver) is UnionType:
            guarded_methods = []
            for cls in toposort_classes(receiver.get_types() - {None}):

                # Create guard checking if receiver is an instance of this class
                guard = self.type_check(args[0], cls, position, ctx)

                # Translate the method call on this particular receiver's class
                method = self._get_method_call(cls, func_name, list(args), arg_types,
                                               list(targets), node, ctx)

                # Translated method call into a block
                block = self.translate_block(method, position, info)

                # Stores guard and translated method call as tuple in a list
                guarded_methods.append((guard, block))

            # Chain list of guard and function call tuples in an if-then-else
            # statement
            return [chain_if_stmts(guarded_methods, self.viper, position, info, ctx)]
        else:
            # Pass-through
            return self._get_method_call(receiver, func_name, args, arg_types,
                                         targets, node, ctx)

    def get_error_var(self, stmt: ast.AST,
                      ctx: Context) -> 'silver.ast.LocalVarRef':
        """
        Returns the error variable of the try-block protecting stmt, otherwise
        the error return variable of the surrounding function, otherwise
        creates a new local variable of type Exception.
        """
        tries = get_surrounding_try_blocks(ctx.actual_function.try_blocks,
                                           stmt)
        if tries:
            err_var = tries[0].get_error_var(self.translator)
            if err_var.sil_name in ctx.var_aliases:
                err_var = ctx.var_aliases[err_var.sil_name]
            return err_var.ref()
        if ctx.actual_function.declared_exceptions:
            return ctx.error_var.ref()
        else:
            new_var = ctx.current_function.create_variable('error',
                ctx.module.global_module.classes['Exception'], self.translator)
            return new_var.ref()

    def var_type_check(self, name: str, type: PythonType,
                       position: 'silver.ast.Position',
                       ctx: Context, inhale_exhale: bool=True) -> Expr:
        """
        Creates an expression checking if the var with the given name
        is of the given type.
        """
        if name in ctx.var_aliases:
            obj_var = ctx.var_aliases[name].ref()
        else:
            obj_var = self.viper.LocalVar(name, self.viper.Ref,
                                          self.no_position(ctx),
                                          self.no_info(ctx))
        return self.type_check(obj_var, type, position, ctx,
                               inhale_exhale=inhale_exhale)

    def create_predicate_access(self, pred_name: str, args: List, perm: Expr,
                                node: ast.AST, ctx: Context) -> Expr:
        """
        Creates a predicate access for the predicate with the given name,
        with the given args and permission.
        """
        pred_acc = self.viper.PredicateAccess(args, pred_name,
                                              self.to_position(node, ctx),
                                              self.no_info(ctx))
        if ctx.perm_factor:
            pos = self.to_position(node, ctx)
            info = self.no_info(ctx)
            perm = self.viper.PermMul(perm, ctx.perm_factor, pos, info)
        pred_acc_pred = self.viper.PredicateAccessPredicate(pred_acc, perm,
            self.to_position(node, ctx), self.no_info(ctx))
        return pred_acc_pred

    def add_handlers_for_inlines(self, ctx: Context) -> List[Stmt]:
        stmts = []
        old_var_valiases = ctx.var_aliases
        old_lbl_aliases = ctx.label_aliases
        for (added_method, var_aliases, lbl_aliases) in ctx.added_handlers:
            ctx.var_aliases = var_aliases
            ctx.label_aliases = lbl_aliases
            ctx.inlined_calls.append(added_method)
            for block in added_method.try_blocks:
                for handler in block.handlers:
                    stmts += self.translate_handler(handler, ctx)
                if block.else_block:
                    stmts += self.translate_handler(block.else_block, ctx)
                if block.finally_block:
                    stmts += self.translate_finally(block, ctx)
            ctx.inlined_calls.remove(added_method)
        ctx.added_handlers = []
        ctx.var_aliases = old_var_valiases
        ctx.label_aliases = old_lbl_aliases
        return stmts

    def _get_string_value(self, string: str) -> int:
        """
        Computes an integer value that uniquely represents the given string.
        """
        result = 0
        for (index, char) in enumerate(string):
            result += pow(256, index) * ord(char)
        return result

    def is_valid_super_call(self, node: ast.Call, container) -> bool:
        """
        Checks if a super() call is valid:
        It must either have no arguments, or otherwise the
        first arg must be a class, the second a reference to self.
        """
        if not node.args:
            return True
        elif len(node.args) == 2:
            target = do_get_target(node.args[0],
                                   container.module.get_included_modules(),
                                   container)
            return (isinstance(target, PythonClass) and
                    isinstance(node.args[1], ast.Name) and
                    (node.args[1].id == next(iter(container.args))))
        else:
            return False

    def get_target(self, node: ast.AST, ctx: Context) -> PythonModule:
        container = ctx.actual_function if ctx.actual_function else ctx.module
        containers = [ctx]
        if ctx.current_class:
            containers.append(ctx.current_class)
        if isinstance(container, (PythonMethod, PythonIOOperation)):
            containers.append(container)
            containers.extend(container.module.get_included_modules())
        else:
            # Assume module
            containers.extend(container.get_included_modules(()))
        result = do_get_target(node, containers, container)
        return result

    def get_fresh_int_lit(self, ctx: Context) -> Expr:
        """
        Returns an integer literal with a fresh value.
        """
        return self.viper.IntLit(ctx.get_fresh_int(), self.no_position(ctx),
                                 self.no_info(ctx))

    def get_unknown_bool(self, ctx: Context) -> Expr:
        """
        Returns an arbitrary but fixed boolean value.
        """
        pos = self.no_position(ctx)
        info = self.no_info(ctx)
        fresh_int = self.get_fresh_int_lit(ctx)
        param = self.viper.LocalVarDecl('i', self.viper.Int, pos, info)
        return self.viper.FuncApp(ARBITRARY_BOOL_FUNC, [fresh_int], pos, info,
                                  self.viper.Bool, [param])



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
from typing import List

from nagini_translation.lib.program_nodes import PythonTryBlock, PythonVar
from nagini_translation.lib.typedefs import Stmt
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.expression import ExpressionTranslator


class SIFExpressionTranslator(ExpressionTranslator):
    """
    Extended AST Version of the ExpressionTranslator.
    """

    def create_exception_catchers(self, var: PythonVar,
                                  try_blocks: List[PythonTryBlock],
                                  call: ast.Call, ctx: Context) -> List[Stmt]:
        if isinstance(var, PythonVar):
            var = var.ref() # the error variable
        position = self.to_position(call, ctx)
        errnotnull = self.viper.NeCmp(var,
                                      self.viper.NullLit(self.no_position(ctx),
                                                         self.no_info(ctx)),
                                      position, self.no_info(ctx))
        raise_stmt = self.viper.Raise(None, self.no_position(ctx), self.no_info(ctx))
        errcheck = self.viper.If(errnotnull, raise_stmt, self.viper.Skip(),
                                 position, self.no_info(ctx))
        return [errcheck]



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Classes for storing obligation translation state."""


from typing import Optional


class ObligationContext:
    """Current state that is related to obligation translation."""

    def __init__(self) -> None:
        self._loop_stack = []

        self.is_translating_posts = False
        """Are we currently translating a postcondition?"""

    @property
    def current_loop_info(self) -> object:
        """Get info of the inner most loop.

        Method type is set to ``object`` to indicate that the result
        should be opaque to all code except obligation implementation.
        """
        assert self._loop_stack
        return self._loop_stack[-1]

    def get_surrounding_loop_info(self) -> Optional[object]:
        """Get info of the loop surrounding the inner most loop.

        Method type is set to ``object`` to indicate that the result
        should be opaque to all code except obligation implementation.
        """
        if len(self._loop_stack) >= 2:
            # Surrounding loop exists only if there are at least two
            # loops in the loop stack.
            return self._loop_stack[-2]
        else:
            return None

    def push_loop_info(self, info: object) -> None:
        """Push loop information to loop stack.

        This method should be called just before a new loop is being
        translated.
        """
        self._loop_stack.append(info)

    def pop_loop_info(self) -> None:
        """Pop loop information from loop stack.

        This method should be called just after a loop was translated.
        """
        self._loop_stack.pop()

    def is_translating_loop(self) -> bool:
        """Return if we currently translating a loop."""
        return bool(self._loop_stack)
